import click
import stimela
from stimela.main import cli
from omegaconf.omegaconf import OmegaConf
import stimela
from datetime import datetime
from enum import Enum


@cli.command(
    help="""With no arguments, lists stimela configuration settings. With KEY=VALUE arguments, changes the configuration file.
            Default is to change whichever configuration file was loaded (local, virtual env, or user-level).
            """,
    short_help="manipulate configuration settings")
@click.argument("settings", nargs=-1, metavar="KEY=VALUE", required=False) 
@click.option("--local", "-l", "save", flag_value='local', 
                help="Save config file locally.")
@click.option("--virtual-env", "-v", "save", flag_value='venv', 
                help="Save config file at the virtual environment level.")
@click.option("--user", "-u", "save", flag_value='user', 
                help="Save config file at the user level.")
def config(settings, save=None):
    log = stimela.logger()
    from stimela.config import CONFIG_LOADED, CONFIG_LOCATIONS

    if CONFIG_LOADED:
        log.info(f"configuration loaded from {CONFIG_LOADED}")
    else:
        log.info(f"no configuration files found, using built-in defaults")

    # print config, if no key=value args specified
    if not settings:
        for key, value in stimela.CONFIG.opts.items():
            if isinstance(value, Enum):
                value = value.name
            print(f"    {key} = {value}")

    # change config, if key=value args specified
    for keyvalue in settings:
        if "=" not in keyvalue:
            log.error(f"invalid config setting '{keyvalue}', KEY=VALUE expected")
            return 2
        key, value = keyvalue.split("=", 1)
        if key not in stimela.CONFIG.opts:
            log.error(f"unknown config key '{key}'")
            return 2
        stimela.CONFIG.opts[key] = value
        print(f"    setting {key} = {value}")

    # save config if changed, or --save given
    if settings or save:
        if save is None:
            output_config = CONFIG_LOADED or CONFIG_LOCATIONS['user']
        else:
            output_config = CONFIG_LOCATIONS[save]
            if not output_config:
                log.error(f"can't write config file at the {save} level")
                return 2
        with open(output_config, "wt") as fp:
            fp.write(f"## Stimela {stimela.__version__} configuration file\n")
            fp.write(f"## Saved on {datetime.now().ctime()}\n\n")
            OmegaConf.save(config=stimela.CONFIG.opts, f=fp)
        log.info(f"wrote configuration to {output_config}")
                


from scabha.basetypes import get_filelikes, File, URI, Directory, MS
from typing import Dict, List, Set, Tuple, Union, Optional
import pytest


@pytest.fixture(scope="module", params=[File, URI, Directory, MS])
def templates(request):
    
    ft = request.param
    
    TEMPLATES = (
        (Tuple, (), set()),
        (Tuple[int, ...], [1, 2], set()),
        (Tuple[ft, ...], ("foo", "bar"), {"foo", "bar"}),
        (Tuple[ft, str], ("foo", "bar"), {"foo"}),
        (Dict[str, int], {"a": 1, "b": 2}, set()),
        (Dict[str, ft], {"a": "foo", "b": "bar"}, {"foo", "bar"}),
        (Dict[ft, str], {"foo": "a", "bar": "b"}, {"foo", "bar"}),
        (List[ft], [], set()),
        (List[int], [1, 2], set()),
        (List[ft], ["foo", "bar"], {"foo", "bar"}),
        (Set[ft], set(), set()),
        (Set[int], {1, 2}, set()),
        (Set[ft], {"foo", "bar"}, {"foo", "bar"}),
        (Union[str, List[ft]], "foo", set()),
        (Union[str, List[ft]], ["foo"], {"foo"}),
        (Union[str, Tuple[ft]], "foo", set()),
        (Union[str, Tuple[ft]], ("foo",), {"foo"}),
        (Optional[ft], None, set()),
        (Optional[ft], "foo", {"foo"}),
        (Optional[Union[ft, int]], 1, set()),
        (Optional[Union[ft, int]], "foo", {"foo"}),
        (Dict[str, Tuple[ft, str]], {"a": ("foo", "bar")}, {"foo"})
    )

    return TEMPLATES


def test_get_filelikes(templates):

    for dt, v, res in templates:
        assert get_filelikes(dt, v) == res, f"Failed for dtype {dt} and value {v}."



import os

from typing import Dict, List, Any, Dict, Tuple
from stimela.kitchen.cab import Cab, Parameter
from scabha.exceptions import SchemaError
from stimela.exceptions import BackendError
from scabha.basetypes import File, Directory, MS, URI, get_filelikes

## commenting out for now -- will need to fix when we reactive the kube backend (and have tests for it)

def resolve_required_mounts(mounts: Dict[str, bool],
                            params: Dict[str, Any], 
                            inputs: Dict[str, Parameter], 
                            outputs: Dict[str, Parameter],
                            remappings: Dict[str, str] = {}
                            ):
    # helper function to accumulate list of target paths to be mounted
    def add_target(param_name, path, must_exist, readwrite):
        if not os.path.isdir(path):
            path = os.path.dirname(path)
        # if file doesn't exit, bind parent or throw error
        if not os.path.exists(path):
            if must_exist:
                raise SchemaError(f"parameter '{param_name}': path '{path}' does not exist")
            path = os.path.dirname(path)
        path = path.rstrip("/")
        # check for remapping -- this can happen if bind_dirs already mounts this container path to a different 
        # host path, in which case we have a conflict 
        if path in remappings:
            raise BackendError(f"{param_name}: {path} already bound to a different host path")
        # insert into mounts (with identical target path if not already present, and possibly upgrading RO to RW)
        mounts[path] = mounts.get(path, False) or readwrite

    # go through parameters and accumulate target paths
    for name, value in params.items():
        schema = inputs.get(name) or outputs.get(name)
        if schema is None:
            raise SchemaError(f"parameter {name} not in defined inputs or outputs for this cab. This should have been caught by validation earlier!")

        files = get_filelikes(schema._dtype, value)

        if not files:
            continue

        must_exist = schema.must_exist and name in inputs 
        readwrite = schema.writable or name in outputs

        for path in files:
            uri = URI(path)
            if uri.remote:
                continue
            path = uri.path
            path = os.path.abspath(path).rstrip("/")
            realpath = os.path.abspath(os.path.realpath(path))
            add_target(name, realpath, must_exist=must_exist, readwrite=readwrite)
            add_target(name, path, must_exist=must_exist, readwrite=readwrite)
            # check if parent directory access is required
            if schema.path_policies.access_parent or schema.path_policies.write_parent:
                add_target(name, os.path.dirname(path), must_exist=True, 
                           readwrite=schema.path_policies.write_parent)
                add_target(name, os.path.dirname(realpath), must_exist=True, 
                           readwrite=schema.path_policies.write_parent)
    
    # now, for any mount that has a symlink in the path, add the symlink target to mounts
    for path, readwrite in list(mounts.items()):
        if path.startswith("::"):
            continue
        while path != "/":
            if os.path.islink(path):
                chain = [path]
                while os.path.islink(path):
                    path = os.readlink(path)
                    # Check if the path is absolute; if not, resolve it relative to the directory of the previous link.
                    if not os.path.isabs(path):
                        path = os.path.abspath(os.path.join(os.path.dirname(chain[-1]), path))
                    chain.append(path)
                    add_target(name, path, must_exist=True, readwrite=readwrite)
            path = os.path.dirname(path)

    # now eliminate unnecessary mounts (those that have a parent mount with no lower read/write privileges)
    skip_targets = set()

    for path, readwrite in mounts.items():
        if path.startswith("::"):
            continue
        parent = os.path.dirname(path)
        while parent != "/":
            # if parent already mounted, and is as writeable as us, skip us
            if parent in mounts and mounts[parent] >= readwrite:
                skip_targets.add(path)
                break
            parent = os.path.dirname(parent)

    for path in skip_targets:
        mounts.pop(path)

def resolve_remote_mounts(params: Dict[str, Any], 
                            inputs: Dict[str, Parameter], 
                            outputs: Dict[str, Parameter],
                            cwd: str = "/",
                            mounts: set = set()):

    must_exist_list = set()
    mkdir_list = set()
    remove_if_exists_list = set()
    active_mounts = set()

    # go through parameters and accumulate target paths
    for name, value in params.items():
        schema = inputs.get(name) or outputs.get(name)
        if schema is None:
            raise SchemaError(f"parameter {name} not in defined inputs or outputs for this cab. This should have been caught by validation earlier!")

        dtype = schema._dtype 
        if dtype in (File, Directory, MS):
            files = [value]
        elif dtype in (List[File], List[Directory], List[MS]):
            files = value
        else:
            continue

        # normalize paths, add CWD, and check that they refer to mounted volumes
        checked_files = []
        for path in files:
            if not os.path.isabs(path):
                path = os.path.join(cwd, path)
            path = os.path.normpath(path)
            # check that it is in mounts
            volumes = set(mount for mount in mounts if os.path.commonpath([path, mount]) == mount)
            if not volumes:
                raise BackendError(f"{name}={path} does not refer to mounted volume")
            active_mounts.update(volumes)
            checked_files.append(path)

        # add to lists of checks to be performed
        must_exist = schema.must_exist
        if must_exist is None:
            must_exist = name in inputs

        if must_exist and name in inputs:
            must_exist_list.update(checked_files)
        if schema.mkdir:
            mkdir_list.update([os.path.dirname(path) for path in checked_files])
        # parent directory must exist, in case of outputs
        elif name in outputs:
            must_exist_list.update([os.path.dirname(path) for path in checked_files])
        if schema.path_policies.remove_if_exists:
            remove_if_exists_list.update(checked_files)

    return must_exist_list, mkdir_list, remove_if_exists_list, active_mounts

        

        


import os.path
import importlib
import re
import fnmatch
from collections.abc import Sequence

from omegaconf.omegaconf import OmegaConf, DictConfig, ListConfig
from omegaconf.errors import OmegaConfBaseException
from typing import Any, List, Dict, Optional, OrderedDict, Union, Callable

from .common import *
from .deps import ConfigDependencies, FailRecord


def _lookup_nameseq(name_seq: List[str], source_dict: Dict):
    """Internal helper: looks up nested item ('a', 'b', 'c') in a nested dict

    Parameters
    ----------
    name_seq : List[str]
        sequence of keys to look up
    source_dict : Dict
        nested dict

    Returns
    -------
    Any
        value if found, else None
    """
    source = source_dict
    names = list(name_seq)
    while names:
        source = source.get(names.pop(0), None)
        if source is None:
            return None
    return source        


def _lookup_name(name: str, *sources: List[Dict]):
    """Internal helper: looks up a nested item ("a.b.c") in a list of dicts

    Parameters
    ----------
    name : str
        section name to look up, e.g. "a.b.c"

    Returns
    -------
    Any
        first matching item found

    Raises
    ------
    NameError
        if matching item is not found
    """
    name_seq = name.split(".")
    for source in sources:
        result = _lookup_nameseq(name_seq, source)
        if result is not None:
            return result
    raise ConfigurattError(f"unknown key {name}")


def _flatten_subsections(conf, depth: int = 1, sep: str = "__"):
    """Recursively flattens subsections in a DictConfig (modifying in place)
    A structure such as
        a:
            b: 1
            c: 2
    Becomes
        a__b: 1
        a__c: 2

    Args:
        conf (DictConfig): config to flatten
        depth (int):       depth to which to flatten. Default is 1 level.
        sep (str):         separator to use, default is "__"
    """
    subsections = [(key, value) for key, value in conf.items() if isinstance(value, DictConfig)]
    for name, subsection in subsections:
        pop_conf(conf, name)
        if depth > 1:
            _flatten_subsections(subsection, depth-1, sep)
        for key, value in subsection.items():
            conf[f"{name}{sep}{key}"] = value


def _scrub_subsections(conf: DictConfig, scrubs: Union[str, List[str]]):
    """
    Scrubs named subsections from a config.

    Args:
        conf (DictConfig): config to scrub
        scrubs (Union[str, List[str]]): sections to remove (can include dots to remove nested sections)
    """
    if isinstance(scrubs, str):
        scrubs = [scrubs]
    
    for scrub in scrubs:
        if '.' in scrub:
            name, remainder = scrub.split(".", 1)
        else:
            name, remainder = scrub, None
        # apply name as pattern
        is_pattern = '*' in name or '?' in name
        matches = fnmatch.filter(conf.keys(), name)
        if not matches:
            # if no matches to pattern, it's ok, otherwise raise error
            if is_pattern:
                return
            raise ConfigurattError(f"no entry matching '{name}'")
        # recurse into or remove matching entries
        for key in matches:
            if remainder:
                subconf = conf[key]
                if type(subconf) is DictConfig:
                    _scrub_subsections(subconf, remainder)
                elif not is_pattern:
                    raise ConfigurattError(f"'{name}' does not refer to a subsection")
            else:
                del conf[key]

def resolve_config_refs(conf, pathname: str, location: str, name: str, includes: bool, 
                        use_sources: Optional[List[DictConfig]],
                        use_cache = True, 
                        include_path: Optional[str]=None,
                        include_stack=[]):
    """Resolves cross-references ("_use" and "_include" statements) in config object

    Parameters
    ----------
    conf : OmegaConf object
        input configuration object
    pathname : str
        full path to this config (directory component of that is used for _includes)
    location : str
        location of this configuration section, used for messages
    name : str
        name of this configuration file, used for messages
    includes : bool
        If True, "_include" references will be processed
    use_sources : optional list of OmegaConf objects
        one or more config object(s) in which to look up "_use" references. None to disable _use statements
    include_path (str, optional):
        if set, path to each config file will be included in the section as element 'include_path'
    include_stack (list, optional):
        stack of files from which this one was included. Used to catch recursion.

    Returns
    -------
    Tuple of (conf, dependencies)
    conf : OmegaConf object    
        This may be a new object if a _use key was resolved, or it may be the existing object
    dependencies : ConfigDependencies
        Set of filenames that were _included

    Raises
    ------
    ConfigurattError
        If a _use or _include directive is malformed
    """
    errloc = f"config error at {location or 'top level'} in {name}"
    dependencies = ConfigDependencies()
    # self-referencing enabled if first source is ourselves
    selfrefs =  use_sources and conf is use_sources[0]

    from scabha.configuratt import load, PATH

    if isinstance(conf, DictConfig):
        
        ## NB: perhaps have _use and _include take effect at the point they're inserted?
        ## also add an _all statement to insert a section into all section that follow
        # since _use and _include statements can be nested, keep on processing until all are resolved        
        updated = True
        recurse = 0
        
        while updated:
            updated = False
            # check for infinite recursion
            recurse += 1
            if recurse > 20:
                raise ConfigurattError(f"{errloc}: recursion limit exceeded, check your _use and _include statements")
            # handle _include/_include_post entries
            if includes:
                # helper function: process includes recursively
                def process_include_directive(include_files: List[str], keyword: str, directive: Any, subpath=None):
                    if isinstance(directive, str):
                        include_files.append(directive if subpath is None else f"{subpath}/{directive}")
                    elif isinstance(directive, (tuple, list, ListConfig)):
                        for dir1 in directive:
                            process_include_directive(include_files, keyword, dir1, subpath)
                    elif isinstance(directive, DictConfig):
                        for key, value in directive.items_ex():
                            process_include_directive(include_files, keyword, value, subpath=key if subpath is None else f"{subpath}/{key}")
                    else:
                        raise ConfigurattError(f"{errloc}: {keyword} contains invalid entry of type {type(directive)}")
                
                # helper function: load list of _include or _include_post files, returns accumulated DictConfig
                def load_include_files(keyword):
                    # pop include directive, return if None
                    include_directive = pop_conf(conf, keyword, None)
                    if include_directive is None:
                        return None
                    # get corresponding _scrub or _scrub_post directive
                    scrub = pop_conf(conf, keyword.replace("include", "scrub"), None)
                    if isinstance(scrub, str):
                        scrub = [scrub]

                    include_files = []
                    process_include_directive(include_files, keyword, include_directive)

                    accum_incl_conf = OmegaConf.create()

                    # load includes
                    for incl in include_files:
                        if not incl:
                            raise ConfigurattError(f"{errloc}: empty {keyword} specifier")
                        # check for [flags] at end of specifier
                        match = re.match(r'^(.*)\[(.*)\]$', incl)
                        if match:
                            incl = match.group(1)
                            flags = set([x.strip().lower() for x in match.group(2).split(",")])
                            warn = 'warn' in flags
                            optional = 'optional' in flags
                        else:
                            flags = {}
                            warn = optional = False

                        # helper function -- finds given include file (including trying an implicit .yml or .yaml extension)
                        # returns full name of file if found, else return None if include is optional, else 
                        # adds fail record and raises exception.
                        # If opt=True, this is stronger than optional (no warnings raised)
                        def find_include_file(path: str, opt: bool = False):
                            # if path already has an extension, only try the pathname itself
                            if os.path.splitext(path)[1]:
                                paths = [path]
                            # else try the pathname itself, plus implicit extensions
                            else:
                                paths = [path] + [path + ext for ext in IMPLICIT_EXTENSIONS]
                            # now try all of them and return a matching one if found
                            for path in paths:
                                if os.path.isfile(path):
                                    return path
                            # end of loop with no matching files? Raise error
                            else:
                                if opt:
                                    return None
                                elif optional:
                                    dependencies.add_fail(FailRecord(path, pathname, warn=warn))
                                    if warn:
                                        print(f"Warning: unable to find optional include {path}")
                                    return None
                                raise ConfigurattError(f"{errloc}: {keyword} {path} does not exist")

                        # check for (location)filename.yaml or (location)/filename.yaml style
                        match = re.match(r"^\((.+)\)/?(.+)$", incl)
                        if match:
                            modulename, filename = match.groups()
                            if modulename.startswith("."):
                                filename = os.path.join(os.path.dirname(pathname), modulename, filename)
                                filename = find_include_file(filename)
                                if filename is None:
                                    continue
                            else:
                                try:
                                    mod = importlib.import_module(modulename)
                                except ImportError as exc:
                                    if optional:
                                        dependencies.add_fail(FailRecord(incl, pathname, modulename=modulename, 
                                                                         fname=filename, warn=warn))
                                        if warn:
                                            print(f"Warning: unable to import module for optional include {incl}")
                                        continue
                                    raise ConfigurattError(f"{errloc}: {keyword} {incl}: can't import {modulename} ({exc})")
                                if mod.__file__ is not None:
                                    path = os.path.dirname(mod.__file__)
                                else:
                                    path = getattr(mod, '__path__', None)
                                    if path is None:
                                        if optional:
                                            dependencies.add_fail(FailRecord(incl, pathname, modulename=modulename, 
                                                                             fname=filename, warn=warn))
                                            if warn:
                                                print(f"Warning: unable to resolve path for optional include {incl}, does {modulename} contain __init__.py?")
                                            continue
                                        raise ConfigurattError(f"{errloc}: {keyword} {incl}: can't resolve path for {modulename}, does it contain __init__.py?")
                                    path = path[0]

                                filename = find_include_file(os.path.join(path, filename))
                                if filename is None:
                                    continue
                        # absolute path -- one candidate
                        elif os.path.isabs(incl):
                            filename = find_include_file(incl)
                            if filename is None:
                                continue
                        # relative path -- scan PATH for candidates
                        else:
                            paths = ['.', os.path.dirname(pathname)] + PATH
                            candidates = [os.path.join(p, incl) for p in paths] 
                            for filename in candidates:
                                filename = find_include_file(filename, opt=True)
                                if filename is not None:
                                    break
                            # none found in candidates -- process error
                            else:
                                if optional:
                                    dependencies.add_fail(FailRecord(incl, pathname, warn=warn))
                                    if warn:
                                        print(f"Warning: unable to find optional include {incl}")
                                    continue
                                raise ConfigurattError(f"{errloc}: {keyword} {incl} not found in {':'.join(paths)}")

                        # check for recursion
                        for path in include_stack:
                            if os.path.samefile(path, filename):
                                raise ConfigurattError(f"{errloc}: {filename} is included recursively")
                        # load included file
                        incl_conf, deps = load(filename, location=location, 
                                            name=f"{filename}, included from {name}",
                                            includes=True, 
                                            include_stack=include_stack,
                                            use_cache=use_cache,
                                            use_sources=None)   # do not expand _use statements in included files, this is done below

                        dependencies.update(deps)
                        if include_path is not None:
                            incl_conf[include_path] = filename

                        # accumulate included config so that later includes override earlier ones
                        accum_incl_conf = OmegaConf.unsafe_merge(accum_incl_conf, incl_conf)

                    if scrub:
                        try:
                            _scrub_subsections(accum_incl_conf, scrub)
                        except ConfigurattError as exc:
                            raise ConfigurattError(f"{errloc}: error scrubbing {', '.join(scrub)}", exc)
                    
                    return accum_incl_conf

                accum_pre = load_include_files("_include")
                accum_post = load_include_files("_include_post")

                # merge: our section overrides anything that has been included
                conf = OmegaConf.unsafe_merge(accum_pre or {}, conf, accum_post or {})
                if accum_pre or accum_post:
                    updated = True
                if selfrefs:
                    use_sources[0] = conf

            # handle _use entries
            if use_sources is not None:
                def load_use_sections(keyword):
                    merge_sections = pop_conf(conf, keyword, None)
                    if merge_sections is None:
                        return None
                    scrub = pop_conf(conf, keyword.replace("use", "scrub"), None)
                    if type(merge_sections) is str:
                        merge_sections = [merge_sections]
                    elif not isinstance(merge_sections, Sequence):
                        raise TypeError(f"invalid {name}.{keyword} directive of type {type(merge_sections)}")
                    if len(merge_sections):
                        # convert to actual sections
                        merge_sections = [_lookup_name(name, *use_sources) for name in merge_sections]
                        # merge them all together
                        base = merge_sections[0].copy()
                        base.merge_with(*merge_sections[1:])
                        # resolve references before flattening
                        base, deps = resolve_config_refs(base, pathname=pathname, name=name, 
                                                location=f"{location}.{keyword}" if location else keyword, 
                                                includes=includes, 
                                                use_sources=use_sources, use_cache=use_cache,
                                                include_path=include_path)
                        dependencies.update(deps)
                        if scrub:
                            try:
                                _scrub_subsections(base, scrub)
                            except ConfigurattError as exc:
                                raise ConfigurattError(f"{errloc}: error scrubbing {', '.join(scrub)}", exc)
                        return base
                    return None
                
                base = load_use_sections("_use")
                if base is not None:
                    base.merge_with(conf)
                    conf = base
                post = load_use_sections("_use_post")
                if post is not None:
                    conf.merge_with(post)
                
                if selfrefs:
                    use_sources[0] = conf

        # recurse into content
        for key, value in conf.items_ex(resolve=False):
            if isinstance(value, (DictConfig, ListConfig)):
                value1, deps = resolve_config_refs(value, pathname=pathname, name=name, 
                                                location=f"{location}.{key}" if location else key, 
                                                includes=includes, 
                                                include_stack=include_stack,
                                                use_sources=use_sources, use_cache=use_cache,
                                                include_path=include_path)
                dependencies.update(deps)
                # reassigning is expensive, so only do it if there was an actual change 
                if value1 is not value:
                    conf[key] = value1
                    
    # recurse into lists
    elif isinstance(conf, ListConfig):
        # recurse in
        for i, value in enumerate(conf._iter_ex(resolve=False)):
            if isinstance(value, (DictConfig, ListConfig)):
                value1, deps = resolve_config_refs(value, pathname=pathname, name=name, 
                                                location=f"{location or ''}[{i}]", 
                                                includes=includes, 
                                                include_stack=include_stack,
                                                use_sources=use_sources, use_cache=use_cache,
                                                include_path=include_path)
                dependencies.update(deps)
                if value1 is not value:
                    conf[i] = value

    return conf, dependencies



from contextlib import contextmanager
from kubernetes import client, config, watch
from kubernetes.client.rest import ApiException
from pprint import pprint
import yaml
import time

from stimela.backends.kube.daskjob import render, create_parser, split_args


    # Define custom resource details
DASKJOB_GROUP = 'kubernetes.dask.org'  # the CRD's group name
DASKJOB_VERSION = 'v1'  # the CRD's version
DASKJOB_PLURAL = 'daskjobs'  # the plural name of the CRD


def find_daskjob_components(daskjob, namespace):
    # pprint(daskjob)
    time.sleep(2)
    v1 = client.CoreV1Api()
    job_name = daskjob["metadata"]["name"]

    runner_pod = v1.list_namespaced_pod(namespace=namespace,
                                        label_selector=f"dask.org/cluster-name={job_name},dask.org/component=job-runner")
    runner_pod_name = runner_pod.items[0].metadata.name

    scheduler_pod = v1.list_namespaced_pod(namespace=namespace, label_selector=f"dask.org/cluster-name={job_name},dask.org/component=scheduler")
    scheduler_pod_name = scheduler_pod.items[0].metadata.name

    worker_pods = v1.list_namespaced_pod(namespace=namespace, label_selector=f"dask.org/cluster-name={job_name},dask.org/component=worker")
    worker_pod_names = [wp.metadata.name for wp in worker_pods.items]


    w = watch.Watch()
    for e in w.stream(v1.read_namespaced_pod_log, name=runner_pod_name, namespace=namespace):
        print(f"Event {e}")
    # response = v1.read_namespaced_pod_log(name=runner_pod_name, namespace=namespace)

    # pprint(response)

    # print(runner_pod_name)
    # print(scheduler_pod_name)
    # print(worker_pod_names)



def daskjob_factory(job_name="sjp", image="ghcr.io/dask/dask", namespace="rarg-test-compute"):
    # CustomObjectsApi is used to handle custom resources
    api_instance = client.CustomObjectsApi()


    arg_list = [
        "--nworkers", "2",
        "--job-name", job_name,
        "--namespace", namespace,
        "--image", image,
        "--cpu-request", "1",
        "--mem-request", "8Gi",
        "--cpu-limit", "1",
        "--mem-limit", "8Gi",
        "--",
        # Job arguments (not scheduler or worker) go here
        "python -c 'from distributed import Client; client = Client(); while True: continue'"
    ]
    args, cmdline = split_args(arg_list)
    args = create_parser().parse_args(args)
    args.cmdline = cmdline
    args.labels = ()
    args.threads_per_worker = 1
    body = render(args)[0] # DaskJob CRD is first in list, Volume stuff + Configmaps follow

    # Create an instance of the CRD
    api_instance.create_namespaced_custom_object(DASKJOB_GROUP, DASKJOB_VERSION, namespace, DASKJOB_PLURAL, body)

    # List all instances of the CRD
    try:
        api_response = api_instance.list_namespaced_custom_object(DASKJOB_GROUP, DASKJOB_VERSION, namespace, DASKJOB_PLURAL)
    except ApiException as e:
        print("Exception when calling CustomObjectsApi->list_namespaced_custom_object: %s\n" % e)
    else:
        #api_response = [api_response] if isinstance(api_response, dict) else api_response

        for instance in api_response["items"]:
            if instance["metadata"]["name"] == job_name:
                find_daskjob_components(instance, namespace)

if __name__ == '__main__':
    # Load configuration
    config.load_kube_config()

    JOBNAME="sjp-test"
    NAMESPACE = "rarg-test-compute"

    try:
        daskjob_factory(job_name=JOBNAME, namespace=NAMESPACE)
    finally:
        api_instance = client.CustomObjectsApi()

        # Delete the instance of the CRD
        api_instance.delete_namespaced_custom_object(DASKJOB_GROUP, DASKJOB_VERSION, NAMESPACE, DASKJOB_PLURAL, JOBNAME)



import  os.path
import  re
import logging
import traceback
import copy
from types import TracebackType
from typing import Optional, OrderedDict, Union, Any
from omegaconf import DictConfig
from scabha.exceptions import ScabhaBaseException, FormattedTraceback
from scabha.substitutions import SubstitutionNS, forgiving_substitutions_from
import rich.progress
import rich.logging
from rich.tree import Tree
from rich import print as rich_print
from rich.markup import escape
from rich.padding import Padding
from rich.syntax import Syntax
from rich.pretty import Pretty
from rich.errors import MarkupError
from warnings import warn

from . import task_stats

CONSOLE_PRINT_OPTIONS = [
    "sep",
    "end",
    "style",
    "justify",
    "overflow",
    "no_wrap",
    "emoji",
    "markup",
    "highlight",
    "width",
    "height",
    "crop",
    "soft_wrap",
    "new_line_start"
]


class FunkyMessage(object):
    """Class representing a message with two versions: funky (with markup), and boring (no markup)"""
    def __init__(self, funky, boring=None, prefix=None, escape_emojis=True):
        self.boring = boring if boring is not None else funky
        if escape_emojis:
            self.funky = re.sub(r":(\w+):", r":[bold][/bold]\1:", funky)
        self.funky = funky
        self.prefix = prefix
    def __str__(self):
        return self.funky
    def __add__(self, other):
        if isinstance(other, FunkyMessage):
            return FunkyMessage(f"{self}{other}", f"{self.boring}{other.boring}")
        else:
            return FunkyMessage(f"{self}{other}", f"{self.boring}{other}")

def defunkify(arg: Union[str, FunkyMessage]):
    return arg.boring if isinstance(arg, FunkyMessage) else arg

class StimelaConsoleHander(rich.logging.RichHandler):
    def __init__(self, console):
        rich.logging.RichHandler.__init__(self, console=console,
                    highlighter=rich.highlighter.NullHighlighter(), 
                    markup=True,
                    show_level=False, show_path=False, show_time=False, keywords=[])
        self._console = console

    def emit(self, record):
        # NOTE(JSKenyon): If a message requires a custom console print,
        # forward all known arguments to the _console.print method.
        if getattr(record, "custom_console_print", False):
            self._console.print(
                record.msg,
                **{k: getattr(record, k) for k in CONSOLE_PRINT_OPTIONS if hasattr(record, k)}
            )
            return
        try:
            rich.logging.RichHandler.emit(self, record)
        # backstop -- message should have been properly markup-escaped
        except MarkupError:
            record.msg = escape(record.msg)
            self._console.print(f"Malformed markup in log message: {record.msg}", markup=False, style="red")
            self._console.print(f"This is a (probably harmless) bug -- but please report", markup=False, style="red")
            try:
                rich.logging.RichHandler.emit(self, record)
            except MarkupError:
                self._console.print(f"Malformed markup after escaping -- this is surely a bug -- please report", markup=False, style="red")

        if hasattr(record, 'console_payload'):
            self._console.print(record.console_payload, highlight=getattr(record, 'console_highlight', None))

class StimelaLogFormatter(logging.Formatter):
    DEBUG_STYLE   = "dim", ""
    WARNING_STYLE = "", "yellow"
    ERROR_STYLE   = "bold", "red"
    CRITICAL_STYLE  = "bold", "red"

    _ansi_escape = re.compile(r'\x1B[@-_][0-?]*[ -/]*[@-~]')

    def __init__(self, boring=False, override_message_attr=None):
        datefmt = "%Y-%m-%d %H:%M:%S"
        if not boring:
            super().__init__("{asctime} {name} [{style}]{levelname}: {message}[/{style}]", datefmt, style="{")
            self._prefix_fmt = logging.Formatter("[{style}]{prefix} {message}[/{style}]", style="{")
        else:
            super().__init__("{asctime} {name} {levelname}: {message}", datefmt, style="{")
            self._prefix_fmt = logging.Formatter("{prefix} {message}", style="{")
        self.boring = boring
        self._override_attr = override_message_attr

    def format(self, record):
        record = copy.copy(record)
        # override message, if a specific override attribute is present
        if self._override_attr is not None and hasattr(record, self._override_attr):
            record.msg = getattr(record, self._override_attr)
        # apply styling
        if not hasattr(record, 'style'):
            if record.levelno <= logging.DEBUG:
                font, color = self.DEBUG_STYLE
            elif record.levelno >= logging.CRITICAL:
                font, color = self.SEVERE_STYLE
            elif record.levelno >= logging.ERROR:
                font, color = self.ERROR_STYLE
            elif record.levelno >= logging.WARNING:
                font, color = self.WARNING_STYLE
            else:
                font = color = ""
            if getattr(record, 'color', None):
                color = record.color.lower()
            if hasattr(record, 'bold') or hasattr(record, 'boldface'):
                if "bold" not in font:
                    font = f"bold {font}"
            setattr(record, 'style', f"{font} {color}".strip() or "normal")
        # in boring mode, make funky messages boring, and strip ANSI codes
        if self.boring:
            record.msg = self._ansi_escape.sub('', defunkify(record.msg))
        # select format based on whether we have prefix or not
        if hasattr(record, 'prefix'):
            if self.boring:
                record.prefix = defunkify(record.prefix)
            return self._prefix_fmt.format(record)
        else:
            return super().format(record)

_logger = None
_log_console_handler = _log_formatter = _log_file_formatter = _log_boring_formatter = _log_colourful_formatter = None

_boring = False

LOG_DIR = '.'

def is_logger_initialized():
    return _logger is not None


def declare_chapter(title: str, **kw):
    if not _boring:
        progress_console.rule(title, **kw)

def apply_style(text: str, style: str):
    if _boring:
        return text
    else:
        return f"[{style}]{text}[/{style}]"


def logger(name="STIMELA", propagate=False, boring=False, loglevel="INFO"):
    """Returns the global Stimela logger (initializing if not already done so, with the given values)"""
    global _logger, _boring
    if _logger is None:
        _logger = logging.getLogger(name)
        _boring = boring
        if type(loglevel) is str:
            loglevel = getattr(logging, loglevel)
        _logger.setLevel(loglevel)
        _logger.propagate = propagate

        global _log_console_handler, _log_formatter, _log_file_formatter, _log_boring_formatter, _log_colourful_formatter
        global progress_console, progress_bar

        _log_file_formatter = StimelaLogFormatter(boring=True, override_message_attr='logfile_message')
        _log_boring_formatter = StimelaLogFormatter(boring=True)
        _log_colourful_formatter = StimelaLogFormatter(boring=False)

        _log_formatter = _log_boring_formatter if boring else _log_colourful_formatter

        progress_bar, progress_console = task_stats.init_progress_bar(boring=boring)

        _log_console_handler = StimelaConsoleHander(console=progress_console)

        _log_console_handler.setFormatter(_log_formatter)
        _log_console_handler.setLevel(loglevel)

        _logger.addHandler(_log_console_handler)
        _logger_console_handlers[_logger.name] = _log_console_handler

        import scabha
        scabha.set_logger(_logger)

    return _logger

_logger_file_handlers = {}
_logger_console_handlers = {}

# keep track of all log files opened
_previous_logfiles = set()


def has_file_logger(log: logging.Logger):
    return log.name in _logger_file_handlers


def disable_file_logger(log: logging.Logger):
    current_logfile, fh = _logger_file_handlers.get(log.name, (None, None))
    if fh is not None:
        fh.close()
        log.removeHandler(fh)
        del _logger_file_handlers[log.name]


class DelayedFileHandler(logging.FileHandler):
    """A version of FileHandler that also handles directory and symlink creation in a delayed way"""
    def __init__(self, logfile, symlink, mode):
        self.symlink, self.logfile = symlink, logfile
        self.is_open = False
        super().__init__(logfile, mode, delay=True)

    def get_logfile_dir(self):
        """Gets name of logfile and ensures the directory exists"""
        if not self.is_open:
            self.is_open = True
            logdir = os.path.dirname(self.logfile)
            if logdir and not os.path.exists(logdir):            
                os.makedirs(logdir)
                if self.symlink:
                    symlink_path = os.path.join(os.path.dirname(logdir.rstrip("/")) or ".", self.symlink)
                    # remove existing symlink
                    if os.path.islink(symlink_path):
                        os.unlink(symlink_path)
                    # Make symlink to logdir. If name exists and is not a symlink, we'll do nothing
                    if not os.path.exists(symlink_path):
                        os.symlink(os.path.basename(logdir), symlink_path)
        return os.path.dirname(self.logfile)

    def emit(self, record):
        self.get_logfile_dir()
        return super().emit(record)

def setup_file_logger(log: logging.Logger, logfile: str, level: Optional[Union[int, str]] = logging.INFO, symlink: Optional[str] = None):
    """Sets up logging to file

    Args:
        log (logging.Logger): Logger object
        logfile (str): logfile. May contain dirname, which will be created as needed.
        level (Optional[Union[int, str]], optional): Logging level, defaults to logging.INFO.
        symlink (Optional[str], optional): if set, and logfile contains a dirname that is created, sets named symlink to point to it
            (This is useful for patterns such as logfile="logs-YYMMDD/logfile.txt", then logs -> logs-YYMMDD)

    Returns:
        [logging.Logger]: logger object
    """
    current_logfile, fh = _logger_file_handlers.get(log.name, (None, None))
    
    # does the logger need a new FileHandler created
    if current_logfile != logfile:
        log.debug(f"will switch to logfile {logfile} (previous was {current_logfile})")
        # remove old FH if so
        if fh is not None:
            fh.close()
            log.removeHandler(fh)
        # if file was previously open, append, else overwrite
        if logfile in _previous_logfiles:
            mode = 'a'
        else:
            mode = 'w'
            _previous_logfiles.add(logfile)
        # create new FH
        fh = DelayedFileHandler(logfile, symlink, mode)
        fh.setFormatter(_log_file_formatter)
        log.addHandler(fh)

        _logger_file_handlers[log.name] = logfile, fh

        # if logging to console, disable propagation from this sub-logger, and add a console handler
        # This ensures that parent loggers that log to files to not get repeated messages
        if _log_console_handler:
            log.propagate = False
            if log.name not in _logger_console_handlers:
                _logger_console_handlers[log.name] = _log_console_handler
                log.addHandler(_log_console_handler)


    # resolve level
    if level is not None:
        if type(level) is str:
            level = getattr(logging, level, logging.INFO)
        fh.setLevel(level)

    return log


def update_file_logger(log: logging.Logger, logopts: DictConfig, nesting: int = 0, subst: Optional[SubstitutionNS] = None, location=[]):
    """Updates logfiles associated with given logger based on option settings

    Args:
        log (logging.Logger):                          Logger object
        nesting (int):                                 nesting level of this logger
        logopts (Union[StimelaLogConfig, DictConfig]): config settings
        subst (Dict[str, Any]):                        dictionary of substitutions for pathnames in logopts
        location (List[str]):                          location of this logger in the hierarchy  

    Returns:
        [type]: [description]
    """

    if logopts.enable and logopts.nest >= nesting:
        path = os.path.join(logopts.dir or ".", logopts.name + logopts.ext)

        if subst is not None:
            with forgiving_substitutions_from(subst, raise_errors=False) as context: 
                path = context.evaluate(path, location=location + ["log"])
                if context.errors:
                    for err in context.errors:
                        log.error(f"bad substitution in log path: {err}")
                    return None

        # substitute non-filename characters for _
        path = re.sub(r'[^a-zA-Z0-9_./-]', '_', path)

        # setup the logger
        setup_file_logger(log, path, level=logopts.level, symlink=logopts.symlink)
    else:
        disable_file_logger(log)
        log.propagate = True


def get_logfile_dir(log: logging.Logger):
    """Returns filename associated with the logger, or None if not logging to file"""
    logfile, fh = _logger_file_handlers.get(log.name, (None, None))
    if logfile is None:
        return None
    return fh.get_logfile_dir()


def log_exception(*errors, severity="error", log=None):
    """Logs one or more error messages or exceptions (unless they are marked as already logged), and 
    pretty-prints them to the console  as appropriate.
    """
    def exc_message(e):
        if isinstance(e, ScabhaBaseException):
            return escape(e.message)
        elif type(e) is str:
            return escape(e)
        else:
            return escape(f"{type(e).__name__}: {e}")

    if severity == "error":
        colour = "bold red"
        message_dispatch = (log or logger()).error
    else:
        colour = "yellow"
        message_dispatch = (log or logger()).warning
    
    trees = []
    do_log = False
    messages = []

    def add_dict(dd, tree):
        for field, value in dd.items():
            if isinstance(value, (dict, OrderedDict, DictConfig)):
                subtree = tree.add(escape(f"{field}:"))
                add_dict(value, subtree)
            else:
                tree.add(escape(f"{field}: {value}"))

    def add_nested(excs, tree):
        for exc in excs:
            if isinstance(exc, Exception):
                subtree = tree.add(f"{exc_message(exc)}")
                # tbtree = subtree.add("Traceback:")
                # for line in traceback.format_exception(exc):
                #     tbtree.add(line)
                if isinstance(exc, ScabhaBaseException) and exc.nested:
                    add_nested(exc.nested, subtree)
            elif type(exc) is TracebackType:
                subtree = tree.add(apply_style("Traceback:", "dim"))
                for line in traceback.format_tb(exc):
                    subtree.add(apply_style(escape(line.rstrip()), "dim"))
            elif type(exc) is FormattedTraceback:
                subtree = tree.add(apply_style("Traceback:", "dim"))
                for line in exc.lines:
                    subtree.add(apply_style(escape(line), "dim"))
            elif isinstance(exc, (dict, OrderedDict, DictConfig)):
                add_dict(exc, tree)
            else:
                tree.add(str(exc))

    has_nesting = False

    for exc in errors:
        if isinstance(exc, ScabhaBaseException):
            messages.append(exc.message)
            if not exc.logged:
                do_log = exc.logged = True
            tree = Tree(exc_message(exc) if _boring else 
                            f"[{colour}]{exc_message(exc)}[/{colour}]", 
                        guide_style="" if _boring else "dim")
            trees.append(tree)
            if exc.nested:
                add_nested(exc.nested, tree)
                has_nesting = True
        else:
            tree = Tree(exc_message(exc) if _boring else 
                            f"[{colour}]{exc_message(exc)}[/{colour}]", 
                        guide_style="" if _boring else "dim")
            trees.append(tree)
            do_log = True
            messages.append(str(exc))

    if do_log:
        message_dispatch(": ".join(messages))

    printfunc = task_stats.progress_bar.console.print if task_stats.progress_bar is not None else rich_print

    if has_nesting:
        declare_chapter("detailed error report follows", style="red")
        for tree in trees:
            printfunc(Padding(tree, pad=(0,0,0,8)))

def log_rich_payload(log: logging.Logger, message: str, payload: Any, 
                     console_payload: Optional[Any] = None, syntax: Optional[str] = None, 
                     level: int = logging.INFO):
    """Logs a message with a rich payload. File logger will get plain text version ("message: \npayload"), 
    console logger will get a rich renderable console_payload object instead, or a Syntax object. 

    Args:
        log (logging.Logger): logger to use
        message (str): message
        payload (Any): payload string
        console_payload (Any): rich renderable object corresponding to the payload. If not set, uses Pretty(payload)
        syntax (str, optional): if set, uses Syntax(payload, syntax) as the console_payload
        level: (int, optional): logging level, defaults to INFO
    """
    if syntax is not None:
        console_payload = Syntax(payload, syntax, padding=(0, 0, 0, 8))
    elif console_payload is None:
        console_payload = Pretty(payload, indent_size=8)
    log.log(level, f"{message}:", 
            extra=dict(logfile_message=f"{message}: \n{str(payload)}",
                        console_payload=console_payload))



import logging, time, json, datetime, os.path, pathlib, secrets, shlex
from typing import Dict, Optional, Any, List, Callable
from dataclasses import fields
from requests import ConnectionError
from urllib3.exceptions import HTTPError
import subprocess
import yaml
import traceback

from omegaconf import OmegaConf, DictConfig, ListConfig

from stimela.utils.xrun_asyncio import dispatch_to_log
from stimela.exceptions import StimelaCabParameterError, StimelaCabRuntimeError, BackendError
from stimela.stimelogging import log_exception, log_rich_payload
from stimela.task_stats import declare_subcommand, declare_subtask, update_process_status
from stimela.backends import StimelaBackendOptions
from stimela.kitchen.cab import Cab

# needs pip install kubernetes dask-kubernetes

from . import get_kube_api, InjectedFileFormatters, session_id, session_user, resource_labels, session_user_info, KubeBackendOptions
from .kube_utils import StatusReporter

from kubernetes.client.rest import ApiException

from .kube_utils import apply_pod_spec
from . import infrastructure, pod_proxy
from stimela.backends.utils import resolve_remote_mounts


def run(cab: Cab, params: Dict[str, Any], fqname: str,
        backend: StimelaBackendOptions,
        log: logging.Logger, subst: Optional[Dict[str, Any]] = None):
    """Runs cab contents

    Args:
        cab (Cab): cab object
        log (logger): logger to use
        subst (Optional[Dict[str, Any]]): Substitution dict for commands etc., if any.

    Returns:
        Any: return value (e.g. exit code) of content
    """

    kube = backend.kube

    args, log_args = cab.flavour.get_arguments(cab, params, subst, check_executable=False)

    log.debug(f"command line is {' '.join(log_args)}")

    cabstat = cab.reset_status()

    command_name = cab.flavour.command_name

    # generate podname
    tmp_name = session_user + "--" + fqname.replace(".", "--").replace("_", "--")
    token_hex = secrets.token_hex(4)
    podname = tmp_name[0:50] + "--" + token_hex
    # K8s don't like uppercase
    podname = podname.lower()

    namespace, kube_api, custom_obj_api = get_kube_api()

    image_name = cab.flavour.get_image_name(cab, backend)
    if not image_name:
        raise BackendError(f"cab '{cab.name}' does not define an image")

    start_time = datetime.datetime.now()
    def elapsed(since=None):
        """Returns string representing elapsed time"""
        return str(datetime.datetime.now() - (since or start_time)).split('.', 1)[0]

    numba_cache_dir = os.path.expanduser("~/.cache/numba")
    pathlib.Path(numba_cache_dir).mkdir(parents=True, exist_ok=True)

    pod_created = dask_job_created = volumes_provisioned = port_forward_proc = None
    bailout_with_exceptions = []

    def k8s_event_handler(event):
        objkind = event.involved_object.kind
        objname = event.involved_object.name
        if event.message.startswith("Error: ErrImagePull"):
            raise BackendError(f"{objkind} '{objname}': failed to pull the image '{image_name}'. Preceding log messages may contain extra information.",
                               event.to_dict())
        if event.reason == "Failed":
            raise BackendError(f"{objkind} '{objname}' reported a 'Failed' event. Preceding log messages may contain extra information.",
                               event.to_dict())
        # if event.reason == "ProvisioningFailed":
        #     raise BackendError(f"{objkind} '{objname}' reported a 'ProvisioningFailed' event. Preceding log messages may contain extra information.",
        #                       event.to_dict())
        # if event.reason == "FailedScheduling":
        #     raise StimelaCabRuntimeError(f"k8s has reported a 'FailedScheduling' event. Preceding log messages may contain extra information.")


    # define debug-print function
    if kube.debug.verbose:
        def dprint(level, message, payload, console_payload: Optional[Any] = None, syntax: Optional[str] = None):
            if level <= kube.debug.verbose:
                log_rich_payload(log, message, payload, console_payload=console_payload, syntax=syntax)
    else:
        def dprint(level, *args, **kw):
            pass

    statrep = StatusReporter(podname=podname, log=log, kube=kube,
                             event_handler=k8s_event_handler)

    if kube.debug.pause_on_start:
        command = "while sleep 600; do echo debug mode still active; done"
    else:
        command = args[0] + " " + " ".join(shlex.quote(x) for x in args[1:])
        if kube.dir:
            command = f"cd {kube.dir}; {command}"

    try:
        with declare_subtask(f"{cab.name}.kube-init", status_reporter=statrep.update):
            log.info(f"using image {image_name}")

            # create pod labels
            pod_labels = dict(stimela_job=podname,
                                stimela_fqname=fqname,
                                stimela_cab=os.path.basename(cab.name),
                                **resource_labels)

            # depending on whether or not a dask cluster is configured, we do either a DaskJob or a regular pod
            if kube.dask_cluster and kube.dask_cluster.enable:
                log.info(f"defining dask job with a cluster of {kube.dask_cluster.num_workers} workers")

                from . import daskjob
                dask_job_name = f"dj-{token_hex}"
                dask_job_spec = daskjob.render(OmegaConf.create(dict(
                    job_name=dask_job_name,
                    labels=pod_labels,
                    namespace=namespace,
                    image=image_name,
                    pull_policy='Always' if kube.always_pull_images else 'IfNotPresent',
                    memory_limit=kube.dask_cluster.memory_limit if kube.dask_cluster.memory_limit is not None
                                    else kube.dask_cluster.worker_pod.memory and kube.dask_cluster.worker_pod.memory.limit,
                    nworkers=kube.dask_cluster.num_workers,
                    threads_per_worker=kube.dask_cluster.threads_per_worker,
                    # cmdline=["/bin/sh", "-c", "while true;do date;sleep 5; done"],
                    service_account=kube.service_account,
                    mount_file=None,
                    environment_variables=kube.env
                )))

                # apply pod type specifications
                if kube.dask_cluster.worker_pod:
                    dprint(1, kube.dask_cluster.worker_pod)
                    dprint(1, dask_job_spec[0]["spec"]["cluster"]["spec"]["worker"]["spec"])
                    dask_job_spec[0]["spec"]["cluster"]["spec"]["worker"]["spec"] = \
                        apply_pod_spec(kube.dask_cluster.worker_pod, dask_job_spec[0]["spec"]["cluster"]["spec"]["worker"]["spec"],
                                                            kube.predefined_pod_specs, log, kind='worker')
                    dprint(1, "Worker pod success")
                if kube.dask_cluster.scheduler_pod:
                    dprint(1, kube.dask_cluster.scheduler_pod)
                    dprint(1, dask_job_spec[0]["spec"]["cluster"]["spec"]["scheduler"]["spec"])
                    dask_job_spec[0]["spec"]["cluster"]["spec"]["scheduler"]["spec"] = \
                        apply_pod_spec(kube.dask_cluster.scheduler_pod, dask_job_spec[0]["spec"]["cluster"]["spec"]["scheduler"]["spec"],
                                                            kube.predefined_pod_specs, log, kind='scheduler')

            else:
                dask_job_spec = dask_job_name = None

            # form up normal pod spec -- either to be run directly, or injected into the dask job
            pod_manifest = dict(
                apiVersion  =  'v1',
                kind        =  'Pod',
                metadata    = dict(name=podname, labels=pod_labels),
            )

            # form up pod spec
            pod = pod_proxy.PodProxy(kube, podname, image_name, command=command, log=log)

            pod_manifest['spec'] = pod.pod_spec

            # add runtime env settings for port forwarding
            if dask_job_spec:
                pod_manifest['spec']['containers'][0]['env'].append(
                        dict(name="DASK_SCHEDULER_ADDRESS",
                        value=f"tcp://{dask_job_name}-scheduler.{namespace}.svc.cluster.local:8786"))

            # add persistent volumes
            # check that mount paths are set before we try to resolve volumes
            for name, pvc in kube.volumes.items():
                if not pvc.mount:
                    raise BackendError(f"volume {name} does not specify a mount path")

            # get the PVCs etc for the volumes
            volumes_initialized = []
            with declare_subcommand("provisioning storage"):
                volumes_provisioned = infrastructure.resolve_volumes(kube, log=log, step_token=token_hex)
                # create volume specs
                for name, pvc in kube.volumes.items():
                    pod_manifest['spec']['volumes'].append(dict(
                        name = name,
                        persistentVolumeClaim = dict(claimName=pvc.name)
                    ))
                    pod.add_volume(name, pvc.mount)
                    # add init commands, if needed
                    pvc0 = infrastructure.active_pvcs[name]
                    # add session init -- this only happens once, if volume is created
                    if not pvc0.initialized:
                        session_init = pvc.init_commands or []
                        if session_init:
                            # if pvc0.owner != session_user:
                            #     log.warning(f"skipping session initialization on PVC '{name}': owned by {pvc0.owner} not {session_user}")
                            # else:
                                pod.add_volume_init(pvc.mount, session_init, root=True)
                                volumes_initialized.append(pvc0)
                    # add step init
                    if pvc.step_init_commands:
                        # if pvc0.owner != session_user:
                        #     log.warning(f"skipping step initialization on PVC '{name}': owned by {pvc0.owner} not {session_user}")
                        # else:
                            pod.add_volume_init(pvc.mount, pvc.step_init_commands, root=False)

                # add to status reporter
                statrep.set_pvcs(kube.volumes)

            # add commands for checking required files
            pod.add_file_check_commands(params, cab)
            # start the threaded status update, since log reading blocks
            pod.start_status_update_thread()

            # start pod and wait for it to come up
            provisioning_deadline = time.time() + (kube.provisioning_timeout or 1e+10)
            if dask_job_spec is None:
                with declare_subcommand("starting pod"):
                    log.info(f"starting pod {podname} to run {command_name}")
                    dprint(1, "pod manifest", pod_manifest)
                    if kube.debug.save_spec:
                        log.info(f"saving pod manifest to {kube.debug.save_spec}")
                        open(kube.debug.save_spec, "wt").write(yaml.dump(pod_manifest))
                    resp = kube_api.create_namespaced_pod(body=pod_manifest, namespace=namespace)
                    dprint(2, "response", resp)
                    pod_created = resp
                    connected = True
                    while pod.check_status():
                        try:
                            resp = kube_api.read_namespaced_pod_status(name=podname, namespace=namespace,
                                                                    _request_timeout=(1, 1))
                        except (ConnectionError, HTTPError) as exc:
                            if connected:
                                log.warn("lost connection to k8s cluster while waiting for the pod to start")
                                log.warn("this is not fatal if the connection eventually resumes")
                                log.warn("use Ctrl+C if you want to give up")
                                connected = statrep.connected = False
                            time.sleep(1)
                            continue
                        phase = resp.status.phase
                        if not connected:
                            log.info("connection resumed", extra=dict(style="green"))
                            connected = statrep.connected = True
                        if phase == 'Running' or phase == 'Succeeded':
                            pod.dispatch_container_logs(kube.capture_logs_style, job=False)
                            break
                        elif phase == 'Failed':
                            pod.dispatch_container_logs(kube.capture_logs_style)
                            raise BackendError("pod startup failed, check logs above")
                        if time.time() >= provisioning_deadline:
                            log.error("timed out waiting for pod to start. The log above may contain more information.")
                            raise BackendError(f"pod failed to start after {kube.provisioning_timeout}s")
                        time.sleep(1)
            # else dask job
            else:
                with declare_subcommand("starting dask job"):
                    log.info(f"starting dask job {dask_job_name} to run {command_name}")
                    # overwrites dask_job job pod spec with one we built up here --
                    # that's ok since there's nothing useful in there (but look out for args.mounts)
                    dask_job_spec[0]["spec"]["job"]["spec"] = pod.pod_spec
                    dask_job_spec[0]["spec"]["job"]["metadata"] = dict(name=podname)
                    # copy job pod volumes to workers
                    volumes = pod.pod_spec['volumes']
                    mounts = pod.pod_spec['containers'][0]['volumeMounts']
                    for name, spec in dask_job_spec[0]["spec"]["cluster"]["spec"].items():
                        spec['spec'].setdefault('volumes', []).extend(volumes)
                        for cont in spec['spec']['containers']:
                            cont.setdefault('volumeMounts', []).extend(mounts)
                    # start the job
                    group, version, plural = 'kubernetes.dask.org', 'v1', 'daskjobs'
                    dprint(1, "daskjob spec", dask_job_spec[0])
                    if kube.debug.save_spec:
                        log.info(f"saving dask job spec to {kube.debug.save_spec}")
                        open(kube.debug.save_spec, "wt").write(yaml.dump(dask_job_spec[0]))
                    resp = custom_obj_api.create_namespaced_custom_object(group, version,
                                            namespace, plural , dask_job_spec[0])
                    dprint(2, "response", resp)
                    dask_job_created = resp
                    job_status = None
                    connected = True
                    # wait for dask job to start up
                    while pod.check_status():
                        try:
                            resp = custom_obj_api.get_namespaced_custom_object_status(group, version,
                                                    namespace, plural, name=dask_job_name, _request_timeout=(1, 1))
                        except (ConnectionError, HTTPError) as exc:
                            if connected:
                                log.warn("lost connection to k8s cluster while waiting for the daskjob to start")
                                log.warn("this is not fatal if the connection eventually resumes")
                                log.warn("use Ctrl+C if you want to give up")
                                connected = statrep.connected = False
                            time.sleep(1)
                            continue
                        if not connected:
                            log.info("connection resumed", extra=dict(style="green"))
                            connected = statrep.connected = True
                        job_status = 'status' in resp and resp['status']['jobStatus']
                        # get podname from job once it's running
                        if job_status:
                            statrep.set_main_status(job_status)
                            if job_status == 'Running' or job_status.startswith('Success'):
                                podname = resp['status']['jobRunnerPodName']
                                statrep.set_pod_name(podname)
                                log.info(f"job running as pod {podname}")
                                pod.name = podname
                                pod.dispatch_container_logs(kube.capture_logs_style, job=False)
                                break
                            elif job_status == 'Failed':
                                pod.dispatch_container_logs(kube.capture_logs_style)
                                raise BackendError("job startup failed, check logs above")

                        if time.time() >= provisioning_deadline:
                            log.error("timed out waiting for dask job to start. The log above may contain more information.")
                            raise BackendError(f"job failed to start after {kube.provisioning_timeout}s")
                        time.sleep(1)

                    # get other pods associated with DaskJob and watch their logs
                    if kube.dask_cluster.capture_logs:
                        pods = kube_api.list_namespaced_pod(namespace=namespace, label_selector=statrep.label_selector)
                        for auxpod in pods.items:
                            if auxpod.metadata.name != podname:
                                pod.start_logging_thread(auxpod.metadata.name, style=kube.dask_cluster.capture_logs_style)

                    # start port forwarding
                    if kube.dask_cluster.forward_dashboard_port:
                        log.info(f"starting port-forward process for http-dashboard to local port {kube.dask_cluster.forward_dashboard_port}")
                        port_forward_proc = subprocess.Popen([kube.kubectl_path,
                            "port-forward", f"service/{dask_job_name}-scheduler",
                            f"{kube.dask_cluster.forward_dashboard_port}:http-dashboard"])

            log.info(f"  pod started after {elapsed()}")

            # if any volumes had session init commands issued, mark them as initialized
            for pvc in volumes_initialized:
                log.info(f"marking PVC {pvc.name} as initialized")
                pvc.initialized = True
                patch_body = dict(metadata=dict(labels=dict(stimela_pvc_initialized="True")))
                kube_api.patch_namespaced_persistent_volume_claim(
                    name=pvc.name,
                    namespace=namespace,
                    body=patch_body)

            # resp = kube_api.read_namespaced_pod(name=podname, namespace=namespace)
            # log.info(f"  read_namespaced_pod {resp}")

            if kube.debug.pause_on_start:
                log.warning("kube.debug.pause_on_start is enabled")
                log.warning(f"to access the pod, run")
                log.warning(f"  $ kubectl exec -it {podname} -- /bin/bash")
                log.warning(f"your command line inside the pod is:")
                if kube.dir:
                    log.warning(f"  $ cd {kube.dir}")
                log.warning(f"  $ {' '.join(args)}")
                log.warning("press Ctrl+C here when done debugging to halt the pod")
            else:
                # do we need to chdir
                log.info(f"running: {command}")

        with declare_subtask(f"{cab.name}.kube-run", status_reporter=statrep.update):
            retcode = None
            connected = True
            last_log_timestamp = None
            seen_logs = set()
            while retcode is None and pod.check_status():
                try:
                    for entry in kube_api.read_namespaced_pod_log(name=podname, namespace=namespace, container="job",
                                follow=True, timestamps=True,
    #                            since_time=last_log_timestamp,
                                _preload_content=False,
                                _request_timeout=(kube.connection_timeout, kube.connection_timeout),
                            ).stream():
                        if not connected:
                            log.info("connection resumed", extra=dict(style="green"))
                            connected = True
                        # log.info(f"got [blue]{entry.decode()}[/blue]")
                        for line in entry.decode().rstrip().split("\n"):
                            if " " in line:
                                timestamp, content = line.split(" ", 1)
                            else:
                                timestamp, content = line, ""
                            key = timestamp, hash(content)
                            last_log_timestamp = timestamp
                            if key in seen_logs:
                                continue
                            seen_logs.add(key)
                            dispatch_to_log(log, content, command_name, "stdout",
                                            output_wrangler=cabstat.apply_wranglers)

                    # check for return code
                    resp = kube_api.read_namespaced_pod_status(name=podname, namespace=namespace)
                    statrep.connected = connected = True
                    contstat = resp.status.container_statuses[0].state
                    waiting = contstat.waiting
                    running = contstat.running
                    terminated = contstat.terminated
                    if waiting:
                        log.info("container state is 'waiting'")
                        dprint(2, "waiting", waiting)
                    elif running:
                        log.info(f"container state is 'running'")
                        dprint(2, "running", running)
                    elif terminated:
                        retcode = terminated.exit_code
                        log.info(f"container state is 'terminated', exit code is {retcode}")
                        dprint(2, "terminated", terminated)
                        break
                except (ConnectionError, HTTPError) as exc:
                    # this could be a real connection error, or maybe the process has gone quiet
                    # so check with the statrep object, since that maintains its own connection status
                    if not statrep.connected:
                        if connected:
                            log.warn("lost connection to k8s cluster: will try to reconnect")
                            log.warn("this is not fatal if the connection eventually resumes")
                            log.warn("use Ctrl+C if you want to give up")
                            connected = False
                    time.sleep(1)

            # check if output marked it as a fail
            if cabstat.success is False:
                log.error(f"declaring '{command_name}' as failed based on its output")

            # if retcode != 0 and not explicitly marked as success, mark as failed
            if retcode and cabstat.success is not True:
                cabstat.declare_failure(f"{command_name} returns error code {retcode} after {elapsed()}")
                if retcode == 137:
                    log.error(f"the pod was killed with an out-of-memory condition. Check your kube.job_pod.memory settings")
            else:
                log.info(f"{command_name} returns exit code {retcode} after {elapsed()}")

            return cabstat

    # handle various failure modes by logging errors appropriately
    except KeyboardInterrupt:
        log.error(f"k8s invocation of {command_name} interrupted with Ctrl+C after {elapsed()}")
        raise StimelaCabRuntimeError(f"{command_name} interrupted with Ctrl+C after {elapsed()}") from None
    except ApiException as exc:
        if exc.body:
            exc = (exc, json.loads(exc.body))
        traceback.print_exc()
        log.error(f"k8s API error after {elapsed()}: {exc}")
        raise BackendError("k8s API error", exc) from None
    except (ConnectionError, HTTPError) as exc:
        log.error(f"k8s connection error after {elapsed()}: {exc}")
        raise BackendError("k8s connection error", exc) from None
    # this drops out as a normal error response
    except StimelaCabRuntimeError as exc:
        log.error(f"cab runtime error after {elapsed()}: {exc}")
        raise
    except BackendError as exc:
        log.error(f"kube backend error after {elapsed()}: {exc}")
        raise
    except Exception as exc:
        log.error(f"k8s invocation of {command_name} failed after {elapsed()}")
        traceback.print_exc()
        raise StimelaCabRuntimeError("kube backend error", exc) from None

    # cleanup
    finally:
        if kube.debug.pause_on_cleanup:
            log.warning("kube.debug.pause_on_cleanup is enabled -- pausing -- press Ctrl+C to proceed")
            try:
                while True:
                    time.sleep(1)
            except KeyboardInterrupt:
                pass
            log.info("proceeding with cleanup")
        statrep.set_event_handler(None)
        try:
            pod.initiate_cleanup()
            # clean up port forwarder
            if port_forward_proc:
                retcode = port_forward_proc.poll()
                if retcode is not None:
                    log.warning(f"kubectl port-forward process has died with code {retcode}")
                    port_forward_proc.wait()
                else:
                    log.info("terminating kubectl port-forward process")
                    port_forward_proc.terminate()
                    try:
                        retcode = port_forward_proc.wait(1)
                    except subprocess.TimeoutExpired:
                        log.warning("kubectl port-forward process hasn't terminated -- killing it")
                        port_forward_proc.kill()
                        try:
                            retcode = port_forward_proc.wait(1)
                        except subprocess.TimeoutExpired:
                            log.warning("kubectl port-forward process refuses to die")
                    if retcode is not None:
                        log.info(f"kubectl port-forward process has exited with code {retcode}")

            # cleean up PVCs
            if volumes_provisioned:
                infrastructure.delete_pvcs(kube, volumes_provisioned, log=log, step=True)

            if podname and pod_created: # or dask_job_created:
                try:
                    update_process_status()
                    log.info(f"deleting pod {podname}")
                    resp = kube_api.delete_namespaced_pod(name=podname, namespace=namespace)
                    log.debug(f"delete_namespaced_pod({podname}): {resp}")
                    # while True:
                    #     resp = kube_api.read_namespaced_pod(name=podname, namespace=namespace)
                    #     log.debug(f"read_namespaced_pod({podname}): {resp}")
                    #     log.info(f"  pod phase is {resp.status.phase} after {elapsed()}")
                    #     time.sleep(.5)
                    update_process_status()
                except ApiException as exc:
                    body = json.loads(exc.body)
                    log_exception(StimelaCabRuntimeError(f"k8s API error while deleting pod {podname}", (exc, body)), severity="warning")
                except Exception as exc:
                    traceback.print_exc()
                    log_exception(StimelaCabRuntimeError(f"error while deleting pod {podname}: {exc}"), severity="warning")
            if dask_job_created:
                try:
                    update_process_status()
                    log.info(f"deleting dask job {dask_job_name}")
                    custom_obj_api.delete_namespaced_custom_object(group, version, namespace, plural, dask_job_name)
                except ApiException as exc:
                    body = json.loads(exc.body)
                    log_exception(StimelaCabRuntimeError(f"k8s API error while deleting dask job {dask_job_name}", (exc, body)), severity="warning")
                except Exception as exc:
                    log_exception(StimelaCabRuntimeError(f"error while deleting dask job {dask_job_name}: {exc}"), severity="warning")
            # wait for aux pod threads
            pod.cleanup()
        except KeyboardInterrupt:
            log.error(f"kube cleanup interrupted with Ctrl+C after {elapsed()}")
            raise StimelaCabRuntimeError(f"{command_name} cleanup interrupted with Ctrl+C after {elapsed()}")
        except Exception as exc:
            log.error(f"kube cleanup failed after {elapsed()}")
            traceback.print_exc()
            raise StimelaCabRuntimeError("kube backend cleanup error", exc)







import os, sys
import click
from omegaconf import OmegaConf
from typing import Dict, List

import stimela
from stimela import logger, log_exception
from stimela.main import cli
from stimela.kitchen.recipe import Recipe
from stimela.exceptions import RecipeValidationError, BackendError

from .run import load_recipe_files

@cli.command("cleanup",
    help="""
    Cleans up backend resources associated with recipe(s).
    """)
@click.argument("items", nargs=-1, metavar="filename.yml...") 
def cleanup(items: List[str] = []):
    
    log = logger()
        
    # load all recipe/config files
    # load config and recipes from all given files
    load_recipe_files(items)

    # now cleanup backend
    backends_list = stimela.CONFIG.opts.backend.select
    if type(backends_list) is str:
        backends_list = [backends_list]
    if not backends_list:
        log.info(f"configuration does not specify any backends, nothing to clean up")
    else:
        log.info(f"invoking cleanup procedure, selected backends: {', '.join(backends_list)}")
        try:
            backend = OmegaConf.to_object(stimela.CONFIG.opts.backend)
            stimela.backends.cleanup_backends(backend, log)
        except BackendError as exc:
            log_exception(exc)
            return 1




from dataclasses import dataclass
import sys
import pytest
from scabha import configuratt, cargo
from omegaconf import OmegaConf
from typing import *
import os

testdir = os.path.dirname(os.path.abspath(__file__))

@dataclass
class SimpleSchema():
    test: Dict[str, cargo.Parameter]

def test_schema():
    path = os.path.join(testdir, "test_schema.yaml")
    conf, _ = configuratt.load(path, use_sources=[])
    conf = OmegaConf.merge(OmegaConf.structured(SimpleSchema), conf)
    OmegaConf.save(conf, sys.stdout)

    print(type(conf.test.a.default))

    obj = OmegaConf.to_object(conf)
    print(obj)


if __name__ == "__main__":
    test_schema()
    


import subprocess
import os
import platform
import getpass
from io import StringIO
from stimela import utils
import json
import stimela
import time
import datetime
import subprocess
import yaml
from shutil import which
from typing import Dict, Union, Optional
from typing import Any, List, Dict, Optional, Union
from dataclasses import dataclass
from omegaconf.omegaconf import MISSING, OmegaConf
from omegaconf.errors import OmegaConfBaseException
from collections import OrderedDict
from scabha.basetypes import EmptyDictDefault, EmptyListDefault

import stimela
from stimela.exceptions import *
from stimela import log_exception

STATUS = VERSION = BINARY = None

def is_available(opts = None):
    global STATUS, VERSION, BINARY
    if STATUS is None:
        BINARY = which("docker")
        if BINARY:
            __version_string = subprocess.check_output([BINARY, "--version"]).decode("utf8")
            STATUS = VERSION = __version_string.strip().split()[-1]
            # if VERSION < "3.0.0":
            #     suffix = ".img"
            # else:
            #     suffix = ".sif"
        else:
            STATUS = "not installed"
            VERSION = None    
            BINARY = None
    return False
    # return VERSION is not None

def get_status():
    return "not implemented"
    # is_available()
    # return STATUS

def is_remote():
    return False


# @dataclass
# class ImageBuildInfo:
#     info: Optional[str] = ""
#     dockerfile: Optional[str] = "Dockerfile"
#     production: Optional[bool] = True          # False can be used to mark test (non-production) images 


# @dataclass
# class StimelaImage:
#     name: str = MISSING
#     info: str = "image description"
#     images: Dict[str, ImageBuildInfo] = MISSING
#     _path: Optional[str] = None   # path to image definition yaml file, if any

#     # optional library of common parameter sets
#     params: Dict[str, Any] = EmptyDictDefault()



# class DockerError(Exception):
#     pass

# from stimela.backends import StimelaImageBuildInfo, StimelaImageInfo

# _available_images: Union[Dict[str,Dict[str, StimelaImageInfo]], None] = None

# def available_images():
#     """Scans system for available stimela images and returns dicitonary of StimelaImageInfo objects.

#     Stimela docker images are identified by a stimela.image.name label.

#     Returns
#     -------
#     dict
#         dictionary: {image_name: {version: image_info}}  
#     """
#     from stimela.main import log
#     global _available_images
#     if _available_images is None:

#         # get list of image IDs which have the right label
#         proc = subprocess.run(["docker", "images",
#                             "--filter", "label=stimela.image.name", 
#                             "--format", "{{.ID}}"], stdout=subprocess.PIPE)
#         _available_images = {}
#         iids = proc.stdout.split()

#         # inspect details of matching IDs
#         if iids:
#             proc = subprocess.run(["docker", "inspect"] + iids, 
#                                 stdout=subprocess.PIPE)
#             # parse output
#             inspect_data = yaml.safe_load(proc.stdout)
#             for num, image_data in enumerate(inspect_data):
#                 iid = image_data.get('Id')
#                 repotags = image_data.get('RepoTags')
#                 if iid is None:
#                     log.warning(f"failed to parse 'docker inspect' output element {num} {repotags}, skipping")
#                     continue
#                 try:
#                     labels = image_data['ContainerConfig']['Labels']
#                     name = labels['stimela.image.name']
#                     version = labels['stimela.image.version']
#                     build = {}
#                     for key in 'stimela_version', 'user', 'host', 'date':
#                         build[key] = labels[f'stimela.build.{key}']
#                 except KeyError as keyerr:
#                     log.warning(f"failed to parse 'docker inspect' output element {num} {repotags}: missing key {keyerr}, skipping")
#                     continue

#                 _available_images.setdefault(name, {})[version] = StimelaImageInfo(name=name, version=version, iid=iid, 
#                                                                                    full_name=repotags[0], build=StimelaImageBuildInfo(**build))
        
#     return _available_images


# def _get_full_name(image: StimelaImage, version:str):
#     """Returns full image name (e.g. "quay.io/stimela/v2-NAME:VERSION")

#     Parameters
#     ----------
#     image : StimelaImage
#         image object
#     version : str
#         version
#     """
#     from stimela.main import CONFIG
#     if CONFIG.opts.registry:
#         basename = f"{CONFIG.opts.registry}/{CONFIG.opts.basename}"
#     else:
#         basename = CONFIG.opts.basename
#     return f"{basename}{image.name}:{version}"


# def build(image: StimelaImage, version: str):
#     """Builds given image + version

#     Parameters
#     ----------
#     image : StimelaImage
#         image object
#     version : str
#         version to be built, must be present in image.images
#     """
#     from stimela.main import log

#     fullname = _get_full_name(image, version)

#     build_info = image.images[version]
#     cwd = os.path.dirname(image.path)
#     dockerfile = os.path.join(cwd, build_info.dockerfile)
#     log.info(f"building {fullname} using {dockerfile}")

#     subprocess.run(["docker", "build", "-t", fullname, "-f", dockerfile,
#                     "--label", f"stimela.image.name={image.name}", 
#                     "--label", f"stimela.image.version={version}", 
#                     "--label", f"stimela.build.stimela_version={stimela.__version__}", 
#                     "--label", f"stimela.build.user={getpass.getuser()}", 
#                     "--label", f"stimela.build.host={platform.node()}", 
#                     "--label", f"stimela.build.date={datetime.datetime.now().ctime()}", 
#                     cwd], check=True)

#     # reset this to force a rescan in available_images()
#     global _available_images
#     _available_images = None


# def push(image: StimelaImage, version: str):
#     """Pushes given image + version to registry

#     Parameters
#     ----------
#     image : StimelaImage
#         image object
#     version : str
#         version to be pushed
#     """
#     from stimela.main import log

#     fullname = _get_full_name(image, version)
#     log.info(f"pushing {fullname}")

#     subprocess.run(["docker", "push", fullname], check=True)
        


# def pull(image, tag=None, force=False):
#     """ pull a docker image """
#     if tag:
#         image = ":".join([image, tag])

#     utils.xrun("docker", ["pull", image])


# def seconds_hms(seconds):
#     return str(datetime.timedelta(seconds=seconds))


# class Container(object):
#     def __init__(self, image, name,
#                  volumes=None, environs=None,
#                  label="", logger=None,
#                  time_out=-1,
#                  workdir=None,
#                  log_container=None,
#                  cabname=None,
#                  runscript=None):
#         """
#         Python wrapper to docker engine tools for managing containers.
#         """

#         self.image = image
#         self.name = name
#         self.cabnane = cabname
#         self.label = label
#         self.volumes = volumes or []
#         self.environs = environs or []
#         self.logger = logger
#         self.status = None
#         self.WORKDIR = workdir
#         self.RUNSCRIPT = runscript
#         self.PID = os.getpid()
#         self.uptime = "00:00:00"
#         self.time_out = time_out
#         self.cont_logger = utils.logger.StimelaLogger(
#             log_container or stimela.LOG_FILE, jtype="docker")

#     def add_volume(self, host, container, perm="rw", noverify=False):

#         if os.path.exists(host) or noverify:
#             if self.logger:
#                 self.logger.debug("Mounting volume [{0}] in container [{1}] at [{2}]".format(
#                     host, self.name, container))
#             host = os.path.abspath(host)
#         else:
#             raise IOError(
#                 "Directory {0} cannot be mounted on container: File doesn't exist".format(host))

#         self.volumes.append(":".join([host, container, perm]))

#     def add_environ(self, key, value):
#         if self.logger:
#             self.logger.debug("Adding environ varaible [{0}={1}] in container {2}".format(
#                 key, value, self.name))
#         self.environs.append("=".join([key, value]))

#     def create(self, *args):

#         if self.volumes:
#             volumes = " -v " + " -v ".join(self.volumes)
#         else:
#             volumes = ""
#         if self.environs:
#             environs = environs = " -e "+" -e ".join(self.environs)
#         else:
#             environs = ""

#         self._print(
#             "Instantiating container [{}]. The container ID is printed below.".format(self.name))
#         utils.xrun("docker create", list(args) + [volumes, environs, "--rm",
#                                                   "-w %s" % (self.WORKDIR),
#                                                   "--name", self.name,
#                                                   self.image,
#                                                   self.RUNSCRIPT or ""], log=self.logger)

#         self.status = "created"

#     def info(self):

#         output = subprocess.check_output(
#             "docker inspect {}".format(self.name), shell=True).decode()
#         output_file = StringIO(output[3:-3])
#         jdict = json.load(output_file)
#         output_file.close()

#         return jdict

#     def get_log(self):
#         stdout = open(self.logfile, 'w')
#         exit_status = subprocess.call("docker logs {0}".format(self.name),
#                                       stdout=stdout, stderr=stdout, shell=True)
#         if exit_status != 0:
#             self.logger.warn(
#                 'Could not log container: {}. Something went wrong durring execution'.format(self.name))
#             output = 'Task was not started.'
#             stdout.write(output)
#         else:
#             output = stdout.read()

#         stdout.close()
#         return output

#     def start(self, output_wrangler=None):
#         running = True
#         tstart = time.time()
#         self.status = "running"

#         self.cont_logger.log_container(self.name)
#         self.cont_logger.write()
#         self._print("Starting container [{0:s}]. Timeout set to {1:d}. The container ID is printed below.".format(
#             self.name, self.time_out))
#         utils.xrun("docker", ["start", "-a", self.name],
#                        timeout=self.time_out,
#                        logfile=self.logfile,
#                        log=self.logger, output_wrangler=output_wrangler,
#                        kill_callback=lambda: utils.xrun("docker", ["kill", self.name]))
#         uptime = seconds_hms(time.time() - tstart)
#         self.uptime = uptime
#         self._print(
#             "Container [{0}] has executed successfully".format(self.name))

#         self._print("Runtime was {0}.".format(uptime))

#         self.status = "exited"

#     def stop(self):
#         dinfo = self.info()
#         status = dinfo["State"]["Status"]
#         killed = False
#         if status in ["running", "paused"]:
#             try:
#                 utils.xrun("docker stop", [self.name])
#             except KeyboardInterrupt("Received terminate signal. Will stop and remove container first"):
#                 killed = True
#         self.status = 'exited'

#         self._print("Container {} has been stopped.".format(self.name))
#         if killed:
#             self.remove()
#             raise KeyboardInterrupt

#     def image_exists(self):
#         """
#             Check if image exists 
#         """
#         image_ids = subprocess.check_output(f"docker images -q {self.image}".split())
#         if image_ids:
#             return True
#         else:
#             return False

#     def remove(self):
#         dinfo = self.info()
#         status = dinfo["State"]["Status"]
#         killed = False
#         if status == "exited":
#             try:
#                 utils.xrun("docker rm", [self.name])
#             except KeyboardInterrupt:
#                 killed = True
#             if killed:
#                 raise KeyboardInterrupt

#         else:
#             raise DockerError(
#                 "Container [{}] has not been stopped, cannot remove".format(self.name))

#         self.cont_logger.remove('containers', self.name)
#         self.cont_logger.write()

#     def _print(self, message):
#         if self.logger:
#             self.logger.info(message)
#         else:
#             print(message)



import re, os.path, json, zlib, codecs, base64, logging
from typing import Optional, Any, Union, Dict
from dataclasses import dataclass

import stimela
from scabha.exceptions import SubstitutionError
from stimela.exceptions import CabValidationError
from stimela.kitchen.cab import Cab
from scabha.cab_utils import CAB_OUTPUT_PREFIX
from stimela.kitchen import wranglers
from scabha.substitutions import substitutions_from

from . import _CallableFlavour, _BaseFlavour


def form_python_function_call(function: str, cab: Cab, params: Dict[str, Any]):
    """
    Helper. Converts a function name and a list of parameters into a string
    representation of a Python function call that can be parsed by the interpreter.
    Uses the cab schema info and policies to decide which parametets to include.

    Args:
        function (str) function name
        cab (Cab): cab definition
        params (Dict[str, Any]): dict of parameters, e.g. {a: 1, b: 'foo'}

    Returns:
        str: function invocation, e.g. "func(a=1, b='foo')"
    """
    arguments = []
    for key, value in cab.filter_input_params(params).items():
        arguments.append(f"{key}={repr(value)}")
    return f"{function}({', '.join(arguments)})"


def format_dict_as_function_call(func_name: Optional[str], d: Dict[str, Any], indent=4):
    """formats dict as a function call"""
    if func_name:
        lines = [f"{func_name}("] 
        comma = ","
    else:
        lines = []
        comma = ''
    pad = ' ' * indent
    for k, v in d.items():
        lines.append(f"{pad}{k}={repr(v)}{comma}")
    # strip trailing comma
    if func_name:
        lines[-1] = lines[-1][:-1] + ")"
    return lines


def get_python_interpreter_args(cab: Cab, subst: Dict[str, Any], virtual_env: Optional[str]=None):
    """
    Helper. Given a cab definition, forms up appropriate argument list to
    invoke the interpreter. Invokes a virtual environment as appropriate.

    Args:
        cab (Cab):              cab definition
        subst (Dict[str, Any]): substitution namespace

    Raises:
        CabValidationError: on errors

    Returns:
        List[str]: [command, arguments, ...] needed to invoke the interpreter
    """
    # get virtual env, if specified
    if virtual_env:
        virtual_env = os.path.expanduser(virtual_env)
        interpreter = f"{virtual_env}/bin/{cab.flavour.interpreter_binary}"
        if not os.path.isfile(interpreter):
            raise CabValidationError(f"{interpreter} doesn't exist")
    else:
        interpreter = cab.flavour.interpreter_binary

    args =  cab.flavour.interpreter_command.format(python=interpreter).split()

    return args


@dataclass
class PythonCallableFlavour(_CallableFlavour):
    """
    Represents a cab flavour that is a Python callable. Cab command field is
    expected to be in the form of [package.]module.function
    """
    kind: str = "python"
    # name of python binary to use  
    interpreter_binary: str = "python"
    # Full command used to launch interpreter. {python} gets substituted for the interpreter path
    interpreter_command: str = "{python} -u"
    # commands run prior to invoking function
    pre_commands: Optional[Dict[str, str]] = None
    # commands run post invoking function
    post_commands: Optional[Dict[str, str]] = None

    def finalize(self, cab: Cab):
        super().finalize(cab)
        # form up outputs handler
        if self.output is not None or self.output_dict:
            self._yield_output = f"print(f'{CAB_OUTPUT_PREFIX}{{json.dumps(_result)}}')"
            pattern = re.compile(f"{CAB_OUTPUT_PREFIX}(.*)")
            if self.output_dict:
                wrangler = wranglers.ParseJSONOutputDict(pattern, "PARSE_JSON")
            else:
                wrangler = wranglers.ParseOutput(pattern, "PARSE_OUTPUT", self.output, '1', "json")
            wrangs = [wrangler]
            if not stimela.VERBOSE:
                wrangs.append(wranglers.Suppress(pattern, "SUPPRESS"))
            cab._wranglers.append((pattern, wrangs))
        else:
            self._yield_output = ""

    def get_image_name(self, cab: Cab, backend: 'stimela.backend.StimelaBackendOptions'):
        from stimela import CONFIG
        from stimela.backends import resolve_image_name
        return resolve_image_name(backend, cab.image or CONFIG.images['default-python'])

    def get_arguments(self, cab: Cab, params: Dict[str, Any], subst: Dict[str, Any],
                      virtual_env: Optional[str]=None, check_executable: bool = True,
                      log: Optional[logging.Logger] = None):
        # substitute command and split into module/function
        with substitutions_from(subst, raise_errors=True) as context:
            try:
                command = context.evaluate(cab.command, location=["command"])
            except Exception as exc:
                raise SubstitutionError(f"error substituting Python callable '{cab.command}'", exc)

        if '.' in command:
            py_module, py_function = cab.command.rsplit('.', 1)
        else:
            raise CabValidationError(f"cab {cab.name}: python flavour requires a command of the form module.function")
        self.command_name = py_function

        # convert inputs into a JSON string
        pass_params = cab.filter_input_params(params)
        pass_params = {key.replace("-","_").replace(".","__"): value for key, value in pass_params.items()}
        params_string = base64.b64encode(
                            zlib.compress(json.dumps(pass_params).encode('ascii'), 2)
                        ).decode('ascii')
        
        # log invocation
        if log:
            log.info(f"preparing function call:", extra=dict(prefix="###", style="dim"))
            for line in format_dict_as_function_call(cab.command, pass_params, indent=4):
                log.info(f"    {line}", extra=dict(prefix="###", style="dim"))

        # form up command string
        if stimela.VERBOSE:
            msg1 = f"""print("## importing {py_module}.{py_function}")"""
            msg2 = f"""print(f"## invoking callable {command}({{repr(_inputs)}}) (as click command) using external interpreter")"""
            msg3 = f"""print(f"## invoking callable {command}({{repr(_inputs)}}) using external interpreter")"""
            msg4 = f"""print("## return value is ", _result)"""
        else:
            msg1 = msg2 = msg3 = msg4 = ""

        pre_command_str = ""
        if self.pre_commands:
            pre_command_str += "\n".join(self.pre_commands.values())
        post_command_str = ""
        if self.post_commands:
            post_command_str += "\n".join(self.post_commands.values())

        code = f"""
import sys, json, zlib, base64
_inputs = json.loads(zlib.decompress(
                        base64.b64decode(sys.argv[1].encode("ascii"))
                    ).decode("ascii"))
sys.path.append('.')
{pre_command_str}
{msg1}
from {py_module} import {py_function}
try:
    from click import Command
except ImportError:
    Command = None
if Command is not None and isinstance({py_function}, Command):
    {msg2}
    {py_function} = {py_function}.callback
else:
    {msg3}
    pass
_result = {py_function}(**_inputs)
{msg4}
{self._yield_output}
{post_command_str}
        """

        args = get_python_interpreter_args(cab, subst, virtual_env=virtual_env)
        return args + ["-c", code, params_string], args + ["-c", "..."]


@dataclass
class PythonCodeFlavour(_BaseFlavour):
    """
    Represents a cab flavour that is inlined Python code. Cab command field is
    Python code. Parameters can be passed in as local variables, or a dict
    """
    kind: str = "python-code"
    # if set to a string, inputs will be passed in as a dict assigned to a variable of that name
    input_dict: Optional[str] = None
    # if True, inputs will be passed in as named variables
    input_vars: bool = True
    # if True, outputs will be collected from named variables
    output_vars: bool = True
    # if True, command will have {}-substitutions done on it
    subst: bool = False
    # name of python binary to use  
    interpreter_binary: str = "python"
    # Full command used to launch interpreter. {python} gets substituted for the interpreter path
    interpreter_command: str = "{python} -u"
    # commands run prior to invoking code
    pre_commands: Optional[Dict[str, str]] = None
    # commands run post invoking code
    post_commands: Optional[Dict[str, str]] = None

    def finalize(self, cab: Cab):
        super().finalize(cab)
        pattern = re.compile(f"{CAB_OUTPUT_PREFIX}(.*)")
        wrangs = [wranglers.ParseJSONOutputDict(pattern, "PARSE_JSON")]
        if not stimela.VERBOSE:
            wrangs.append(wranglers.Suppress(pattern, "SUPPRESS"))
        cab._wranglers.append((pattern, wrangs))
        self.command_name = "[python]"

    def get_image_name(self, cab: Cab, backend: 'stimela.backend.StimelaBackendOptions'):
        from stimela import CONFIG
        from stimela.backends import resolve_image_name
        return resolve_image_name(backend, cab.image or CONFIG.images['default-python'])

    def get_arguments(self, cab: Cab, params: Dict[str, Any], subst: Dict[str, Any],
                      virtual_env: Optional[str]=None, check_executable: bool = True,
                      log: Optional[logging.Logger] = None):
        # do substitutions on command, if necessary
        if self.subst:
            with substitutions_from(subst, raise_errors=True) as context:
                try:
                    command = context.evaluate(cab.command, location=["command"])
                except Exception as exc:
                    raise SubstitutionError("error substituting inline Python code", exc)
        else:
            command = cab.command

        # only pass inputs and named outputs
        pass_params = cab.filter_input_params(params)

        # log invocation
        if log:
            log.info(f"preparing python code invocation with arguments:", extra=dict(prefix="###", style="dim"))
            for line in format_dict_as_function_call("", pass_params, indent=4):
                log.info(f"{line}", extra=dict(prefix="###", style="dim"))

        # form up code to parse params from JSON string that will be given as sys.argv[1]
        params_arg = json.dumps(pass_params)
        inp_dict = self.input_dict or "_params"

        pre_command_str = f"""
import sys, json
{inp_dict} = json.loads(sys.argv[1])
"""
        if self.pre_commands:
            pre_command_str += "\n".join(self.pre_commands.values())

        if self.input_vars:
            for name in pass_params:
                var_name = name.replace("-", "_").replace(".", "__")
                pre_command_str += f"""{var_name} = {inp_dict}["{name}"]\n"""

        # form up code to print outputs in JSON
        post_command_str = ""
        pass_outputs = [name for name, schema in cab.outputs.items()
                        if not schema.is_named_output and not schema.implicit]
        if pass_outputs:
            post_command_str += "from scabha.cab_utils import yield_output\n"
            if self.output_vars:
                for name in pass_outputs:
                    var_name = name.replace("-", "_").replace(".", "__")
                    post_command_str += f"yield_output(**{{'{name}': {var_name}}})\n"
        
        if self.post_commands:
            post_command_str += "\n".join(self.post_commands.values())

        # form up interpreter invocation
        args = get_python_interpreter_args(cab, subst, virtual_env=virtual_env)
        return args + ["-c", pre_command_str + command + post_command_str, params_arg], args + ["-c", "..."]



import dataclasses
import os
import os.path
import yaml
import re
import keyword
import pathlib
from typing import *
from collections import OrderedDict

import pydantic
import pydantic.dataclasses

from omegaconf import OmegaConf, ListConfig, DictConfig

from scabha.basetypes import Unresolved
from .exceptions import Error, ParameterValidationError, SchemaError, SubstitutionErrorList
from .substitutions import SubstitutionNS, substitutions_from
from .basetypes import URI, File, Directory, MS, UNSET
from .evaluator import Evaluator

def join_quote(values):
    return "'" + "', '".join(values) + "'" if values else ""


def evaluate_and_substitute_object(obj: Any,  
                                    subst: SubstitutionNS, 
                                    recursion_level: int = 1,
                                    location: List[str] = []):
    with substitutions_from(subst, raise_errors=True) as context:
        evaltor = Evaluator(subst, context, location=location, allow_unresolved=False)
        return evaltor.evaluate_object(obj, raise_substitution_errors=True, recursion_level=recursion_level)


def evaluate_and_substitute(inputs: Dict[str, Any], 
                            subst: SubstitutionNS, 
                            corresponding_ns: SubstitutionNS,
                            defaults: Dict[str, Any] = {},
                            ignore_subst_errors: bool = False, 
                            location: List[str] = []):
    with substitutions_from(subst, raise_errors=True) as context:
        evaltor = Evaluator(subst, context, location=location, allow_unresolved=True)
        inputs = evaltor.evaluate_dict(inputs, corresponding_ns=corresponding_ns, defaults=defaults,
                                        raise_substitution_errors=False)
        # collect errors
        if not ignore_subst_errors:
            errors = []
            for value in inputs.values():
                if type(value) is Unresolved:
                    errors += value.errors
            # check for substitution errors
            if errors:
                raise SubstitutionErrorList("unresolved {}-substitutions", errors)
    return inputs


def validate_parameters(params: Dict[str, Any], schemas: Dict[str, Any], 
                        defaults: Optional[Dict[str, Any]] = None,
                        subst: Optional[SubstitutionNS] = None,
                        fqname: str = "",
                        check_unknowns=True,    
                        check_required=True,
                        check_inputs_exist=True,
                        check_outputs_exist=True,
                        create_dirs=False,
                        ignore_subst_errors=False,
                        ) -> Dict[str, Any]:
    """Validates a dict of parameter values against a given schema 

    Args:
        params (Dict[str, Any]):   map of input parameter values
        schemas (Dict[str, Any]):   map of parameter names to schemas. Each schema must contain a dtype field and a choices field.
        defaults (Dict[str, Any], optional): dictionary of default values to be used when a value is missing
        subst (SubsititionNS, optional): namespace to do {}-substitutions on parameter values
        fqname: fully-qualified name of the parameter set (e.g. "recipe_name.step_name"), used in error messages. If not given,
                errors will report parameter names only

        check_unknowns (bool): if True, unknown parameters (not in schema) raise an error
        check_required (bool): if True, missing parameters with required=True will raise an error
        check_inputs_exist (bool): if True, input files with must_exist={None,True} in schema must exist, or will raise an error. 
                                   If False, doesn't check.
        check_outputs_exist (bool): if True, output files with must_exist=True in schema must exist, or will raise an error. 
                                   If False, doesn't check.
        create_dirs (bool): if True, non-existing directories in filenames (and parameters with mkdir=True in schema) 
                            will be created.
        ignore_subst_errors (bool): if True, substitution errors will be ignored


    Raises:
        ParameterValidationError: parameter fails validation
        SchemaError: bad schema
        SubstitutionErrorList: list of substitution errors, if they occur

    Returns:
        Dict[str, Any]: validated dict of parameters

    TODO:
        add options to propagate all errors out (as values of type Error) in place of exceptions?
    """
    # define function for converting parameter name into "fully-qualified" name
    if fqname:
        mkname = lambda name: f"{fqname}.{name}"
    else:
        mkname = lambda name: name

    # check for unknowns
    if check_unknowns:
        for name in params:
            if name not in schemas:
                raise ParameterValidationError(f"unknown parameter '{mkname(name)}'")
    
    # only evaluate the subset for which we have schemas
    inputs = OrderedDict((name, value) for name, value in params.items() if name in schemas)
    
    # build dict of all defaults 
    all_defaults = {name: schema.default for name, schema in schemas.items() if schema.default is not UNSET and schema.default != "UNSET"}
    if defaults:
        all_defaults.update(**{name: value for name, value in defaults.items() if name in schemas})

    # update missing inputs from defaults
    inputs.update(**{name: value for name, value in all_defaults.items() if name not in inputs})

    # update implicit values
    #  

    # perform substitution
    if subst is not None:
        inputs = evaluate_and_substitute(inputs, subst, subst.current, defaults=all_defaults, 
                                        ignore_subst_errors=ignore_subst_errors, 
                                        location=[fqname])

    # split inputs into unresolved substitutions, and proper inputs
    unresolved = {name: value for name, value in inputs.items() if isinstance(value, Unresolved)}
    inputs = {name: value for name, value in inputs.items() if not isinstance(value, Unresolved)}

    # check that required args are present
    if check_required:
        missing = [mkname(name) for name, schema in schemas.items() 
                    if schema.required and inputs.get(name) is UNSET and name not in unresolved]
        if missing:
            raise ParameterValidationError(f"missing required parameters: {join_quote(missing)}")

    # create dataclass from parameter schema
    validated = {}
    fields = []

    # maps parameter names to/from field names. Fields have "_" not "-"
    name2field = {}
    field2name = {}

    for name, schema in schemas.items():
        value = inputs.get(name, UNSET)
        if value is not UNSET:
            # sanitize name: dataclass won't take hyphens or periods
            fldname = orig_fldname = re.sub("\\W", "_", name)
            # avoid Python keywords and clashes with other field names by adding _x as needed
            num = 0
            while keyword.iskeyword(fldname) or \
                    (hasattr(keyword, 'issoftkeyword') and keyword.issoftkeyword(fldname)) or \
                    fldname in field2name:
                fldname += f"{orig_fldname}_{num}"
                num += 1
            # add to mapping
            field2name[fldname] = name
            name2field[name] = fldname

            fields.append((fldname, schema._dtype))
            
            # OmegaConf dicts/lists need to be converted to standard containers for pydantic to take them
            if isinstance(value, (ListConfig, DictConfig)):
                inputs[name] = OmegaConf.to_container(value)

    dcls = dataclasses.make_dataclass("Parameters", fields)

    # convert this to a pydantic dataclass which does validation
    pcls = pydantic.dataclasses.dataclass(dcls)

    # check Files etc. 
    for name, value in list(inputs.items()):
        # get schema from those that need validation, skip if not in schemas
        schema = schemas.get(name)
        if schema is None:
            continue
        # skip errors
        if value is UNSET or isinstance(value, Error):
            continue
        dtype = schema._dtype

        # must this file exist? Schema may force this check, otherwise follow the default check_exist policy
        if schema.is_input:
            must_exist = check_inputs_exist and schema.must_exist is not False
        elif schema.is_output:
            must_exist = check_outputs_exist and schema.must_exist

        if schema.is_file_type or schema.is_file_list_type:
            # match to existing file(s)
            if type(value) is str:
                # try to interpret string as a formatted list (a list substituted in would come out like that)
                try:
                    files = yaml.safe_load(value)
                    if type(files) is not list:
                        files = [value]
                except Exception as exc:
                    files = [value]
            elif isinstance(value, (list, tuple)):
                files = value
            else:
                raise ParameterValidationError(f"'{mkname(name)}={value}': invalid type '{type(value)}'")
            # convert to appropriate type 
            files = [URI(f) for f in files]
          
            # check for existence of all files in list, if needed
            if must_exist: 
                if not files:
                    raise ParameterValidationError(f"'{mkname(name)}': file(s) don't exist")
                not_exists = [uri.path for uri in files 
                              if not uri.remote and not os.path.exists(uri.path)]
                if not_exists:
                    raise ParameterValidationError(f"'{mkname(name)}': {','.join(not_exists)} doesn't exist")

            # check for single file/dir
            if schema.is_file_type:
                if len(files) > 1:
                    raise ParameterValidationError(f"'{mkname(name)}': multiple files given ({value})")
                # no files? must_exist was checked above, so return empty filename
                elif not files:
                    inputs[name] = "" 
                # else one file/dir as expected, check it                   
                else:
                    # check that files are files and dirs are dirs
                    uri = files[0]
                    if not uri.remote and os.path.exists(uri.path):
                        if dtype == File:
                            if not os.path.isfile(uri.path):
                                raise ParameterValidationError(f"'{mkname(name)}': {uri} is not a regular file")
                        elif dtype == Directory or dtype == MS:
                            if not os.path.isdir(uri.path):
                                raise ParameterValidationError(f"'{mkname(name)}': {uri} is not a directory")
                    inputs[name] = str(uri)
            # else make list
            else:
                # check that files are files and dirs are dirs
                if dtype == List[File]:
                    if not all(os.path.isfile(uri.path) for uri in files 
                               if not uri.remote and os.path.exists(uri.path)):
                        raise ParameterValidationError(f"{mkname(name)}: {value} contains non-files")
                elif dtype == List[Directory] or dtype == List[MS]:
                    if not all(os.path.isdir(uri.path) for uri in files 
                               if not uri.remote and os.path.exists(uri.path)):
                        raise ParameterValidationError(f"{mkname(name)}: {value} contains non-directories")
                inputs[name] = list(map(str, files))

    # validate
    try:   
        validated = pcls(**{name2field[name]: value for name, value in inputs.items() if name in schemas and value is not UNSET})
    except pydantic.ValidationError as exc:
        errors = []
        for err  in exc.errors():
            loc = '.'.join([field2name.get(x, str(x)) for x in err['loc']])
            if loc in inputs:
                errors.append(ParameterValidationError(f"{loc} = {inputs[loc]}: {err['msg']}")) 
            else:
                errors.append(ParameterValidationError(f"{loc}: {err['msg']}")) 
        raise ParameterValidationError(f"{len(errors)} parameter(s) failed validation:", errors)

    validated = {field2name[fld]: value for fld, value in dataclasses.asdict(validated).items()}

    # check choice-type parameters
    for name, value in validated.items():
        if value is None:
            continue
        schema = schemas[name]
        if schema.choices and value not in schema.choices:
            raise ParameterValidationError(f"{mkname(name)}: invalid value '{value}'")
        if schema.element_choices:
            listval = value if isinstance(value, (list, tuple, ListConfig)) else [value]
            for elem in listval:
                if elem not in schema.element_choices:
                    raise ParameterValidationError(f"{mkname(name)}: invalid list element '{elem}'")

    # check for mkdir directives
    if create_dirs:
        for name, value in validated.items():
            schema = schemas[name]
            if schema.is_output and (schema.mkdir or schema.path_policies.mkdir_parent):
                if schema.is_file_type:
                    files = [URI(value)]
                elif schema.is_file_list_type:
                    files = map(URI, value)
                else:
                    continue
                for uri in files:
                    if not uri.remote:
                        path = pathlib.Path(uri.path)
                        # Directory-type outputs 
                        if schema.mkdir and (schema._dtype == Directory or schema._dtype == List[Directory]):
                            path.mkdir(parents=True, exist_ok=True)
                        elif schema.path_policies.mkdir_parent:
                            path.parent.mkdir(parents=True, exist_ok=True)

    # add in unresolved values
    validated.update(**unresolved)

    return validated



import re, logging, json, yaml
from typing import Any, List, Dict, Optional, Union
from omegaconf import ListConfig


from scabha.cargo import ListOrString
from stimela.exceptions import CabValidationError, StimelaCabOutputError, \
                    StimelaCabRuntimeError
from stimela.stimelogging import FunkyMessage

# wranglers specified as a single string, or a list
WranglerSpecList = ListOrString

# defined just for type hinting below
CabStatus = "stimela.kitchen.cab.Cab.RuntimeStatus"


def create(regex: re.Pattern, spec: str):
    """Creates wrangler object from a regex and a specification string, using the spec
    string to look up a wrangler class below.

    Args:
        regex (re.Pattern): pattern in output which wrangler will match
        spec (str): specification string

    Raises:
        CabValidationError: if the specification is incorrect

    Returns:
        wrangler: a wrangler object
    """
    for specifier, wrangler_class in all_wranglers.items():
        match = re.fullmatch(specifier, spec)
        if match:
            return wrangler_class(regex, spec, **match.groupdict())
    raise CabValidationError(f"'{regex.pattern}': '{spec}' is not a valid wrangler specifier")


def create_list(pattern: Union[str, re.Pattern], spec_list: WranglerSpecList):
    """Creates list of wranglers based on list of specifiers.

    Args:
        pattern (Union[str, re.Pattern]): pattern in output which wrangler will match
        spec_list (WranglerSpecList): list of wrangler specs

    Raises:
        CabValidationError: if a specification is incorrect, or pattern is not a valid regex

    Returns:
        (re.Pattern, list): compiled regex pattern, plus list of wrangler objects
    """
    if type(pattern) is str:
        try:
            regex = re.compile(pattern)
        except Exception as exc:
            raise CabValidationError(f"wrangler pattern '{pattern}' is not a valid regular expression", exc)
    else:
        regex = pattern
        pattern = regex.pattern
    
    if type(spec_list) is str:
        spec_list = [spec_list]
    if not isinstance(spec_list, (list, tuple, ListConfig)):
        raise CabValidationError(f"wrangler entry '{pattern}': expected list of wranglers")

    return regex, [create(regex, spec) for spec in spec_list]


class _BaseWrangler(object):
    """Abstract base class for wrangler action classes.
    
    Each action class has a static 'specifier' attribute, which is a pattern used to match (and parse)
    the specifier for that action. The pattern may contain named ()-groups (see e.g. Replace below),
    note that the create() function above maps these to keyword arguments of the constructor.
    """

    specifier = None

    def __init__(self, regex: re.Pattern, spec: str, **kw):
        """Action class constructor. Copies all keywords into object attributes.

        Args:
            regex (re.Pattern): pattern object which triggers this action
            spec (str):         original action specifier string 
            **kw (Any):         converted to object attributes
        """
        self.regex, self.spec = regex, spec
        for key, value in kw.items():
            setattr(self, key, value)

    def apply(self, cabstat: CabStatus, output: str, match: re.Match):
        """Abstract method. Applies the wrangler to a line of cab output.

        Args:
            cabstat (CabStatus): cab runtime status object, which may be updated based on output
            output (str): line of output
            match (re.Match): match object resulting from pattern matching
        
        Returns:
            output (str or None), severity (int or None): [possibly] modified output string, or None to suppress
                                                          modified log severity level, or None to not modify
        """
        raise RuntimeError("derived class does not implement an apply() method")

class Replace(_BaseWrangler):
    """
    This wrangler will replace the matching pattern with the given string. Specified as REPLACE:replacement.
    Uses re.sub() internally, so look that up for more complex usage.
    """ 

    specifier = "REPLACE:(?P<replace>.*)"
    
    def apply(self, cabstat: CabStatus, output: str, match: re.Match):
        return self.regex.sub(self.replace, output), None

class Highlight(_BaseWrangler):
    """
    This wrangler will highlight the matching pattern using the given rich markup. Specified as REPLACE:replacement.
    Uses re.sub() internally, so look that up for more complex usage.
    """
    specifier = "HIGHLIGHT:(?P<style>.*)"
    
    def apply(self, cabstat: CabStatus, output: str, match: re.Match):
        return self.regex.sub(f"[{self.style}]\\g<0>[/{self.style}]", output), None


class ChangeSeverity(_BaseWrangler):
    """
    This wrangler will cause the output to be reported at the given severity level (instead of the
    normal INFO level). Specified as SEVERITY:level, where level is one of the symbolic levels
    defined in the standard logging module.
    """
    specifier = "SEVERITY:(?P<severity>ERROR|WARNING|INFO|DEBUG|CRITICAL|FATAL)"
    
    def __init__(self, regex: re.Pattern, spec: str, severity: str):
        super().__init__(regex, spec, severity=severity)
        level = getattr(logging, severity, None)
        if level is None:
            raise CabValidationError(f"wrangler action '{spec}' for '{regex.pattern}': invalid logging level")
        self.severity = level

    def apply(self, cabstat: CabStatus, output: str, match: re.Match):
        return output, self.severity

class Suppress(_BaseWrangler):
    """
    This wrangler will cause the matching output to be suppressed. Specified as simply SUPPRESS.
    """
    specifier = "SUPPRESS"
    def apply(self, cabstat: CabStatus, output: str, match: re.Match):
        return None, None

class DeclareWarning(_BaseWrangler):
    """
    This wrangler will issue a warning once the cab completes. Specified as WARNING:message
    """
    specifier = "WARNING:(?P<message>.*)"
    def apply(self, cabstat: CabStatus, output: str, match: re.Match):
        cabstat.declare_warning(self.message)
        return output, logging.WARNING

class DeclareError(_BaseWrangler):
    """
    This wrangler will mark the cab as failed (even if the exit code is 0), and issue an optional error.
    Specified as ERROR[:message]
    """
    specifier = "ERROR(?::(?P<message>.*))?"
    def apply(self, cabstat: CabStatus, output: str, match: re.Match):
        if self.message:
            message = self.message.format(**match.groupdict())
        else:
            message = f"cab marked as failed based on encountering '{self.regex.pattern}' in output"
        cabstat.declare_failure(StimelaCabRuntimeError(message))
        return FunkyMessage(f"[bold red]{output}[/bold red]", output), logging.ERROR

class DeclareSuccess(_BaseWrangler):
    """
    This wrangler will mark the cab as succeeded (even if the exit code is non 0).
    Specified as DECLARE_SUCCESS.
    """
    specifier = "DECLARE_SUCCESS"
    def apply(self, cabstat: CabStatus, output: str, match: re.Match):
        cabstat.declare_success()
        return FunkyMessage(f"[bold green]{output}[/bold green]", output, 
            prefix=FunkyMessage(":white_check_mark:","+")), None

class ParseOutput(_BaseWrangler):
    """
    This wrangler will parse a named output parameter out of the output. Specified as 
    PARSE_OUTPUT:name:type or PARSE_OUTPUT:name:group:type. 
    In the first case, the matching pattern must have a ()-group with the same name.
    In the second case, a ()-group name (or number) is explicitly specified.
    Type can be an atomic Python type (str, bool, int, float, complex), or 'json' to use
    json.loads(), or 'yaml' to use yaml.safe_load
    """

    loaders = dict(str=str, bool=bool, int=int, float=float, complex=complex, 
        json=json.loads, JSON=json.loads,
        yaml=yaml.safe_load, YAML=yaml.safe_load)
    specifier = f"PARSE_OUTPUT:((?P<name>.*):)?(?P<group>.*):(?P<dtype>{'|'.join(loaders.keys())})"

    def __init__(self, regex: re.Pattern, spec: str, name: Optional[str], group: str, dtype: str):
        super().__init__(regex, spec, name=name)
        self.loader = self.loaders[dtype]
        self.name = name or group
        if group in regex.groupindex:
            self.gid = group
        elif re.fullmatch(r'\d+', group):
            gid = int(group)
            if gid > regex.groups:
                raise CabValidationError(f"wrangler action '{spec}' for '{regex.pattern}': {gid} is not a valid ()-group")
            self.gid = gid
        else:
            raise CabValidationError(f"wrangler action '{spec}' for '{regex.pattern}': {group} is not a valid ()-group")

    def apply(self, cabstat: CabStatus, output: str, match: re.Match):
        value = match[self.gid]
        try:
            value = self.loader(value)
        except Exception as exc:
            cabstat.declare_failure(StimelaCabOutputError(f"error parsing string \"{value}\" for output '{self.name}'", exc))
        cabstat.declare_outputs({self.name: value})
        return output, None

class ParseJSONOutput(_BaseWrangler):
    """
    This wrangler will parse multiple named output parameters (using JSON) out of the output. Specified as 
    PARSE_JSON_OUTPUTS.
    The pattern is expected to have one or more named ()-groups, which are matched to parameters.
    """

    specifier = "PARSE_JSON_OUTPUTS"

    def __init__(self, regex: re.Pattern, spec: str):
        super().__init__(regex, spec)
        self.names = regex.groupindex.keys()
        if not self.names:
            raise CabValidationError(f"wrangler action '{spec}' for '{regex.pattern}': no ()-groups")

    def apply(self, cabstat: CabStatus, output: str, match: re.Match):
        outputs = {}
        for name, value in match.groupdict().items():
            if value is not None:
                try:
                    outputs[name] = json.loads(value)
                except Exception as exc:
                    cabstat.declare_failure(StimelaCabOutputError(f"error parsing string \"{value}\" for output '{name}'", exc))
        cabstat.declare_outputs(outputs)
        return output, None


class ParseJSONOutputDict(_BaseWrangler):
    """
    This wrangler will parse a dict of output parameters (using JSON) out of the output. Specified as 
    PARSE_JSON_OUTPUT_DICT.
    The first ()-group from the pattern will be parsed.
    """

    specifier = "PARSE_JSON_OUTPUT_DICT"

    def __init__(self, regex: re.Pattern, spec: str):
        super().__init__(regex, spec)
        if regex.groups < 1:
            raise CabValidationError(f"wrangler action '{spec}' for '{regex.pattern}': no ()-groups")

    def apply(self, cabstat: CabStatus, output: str, match: re.Match):
        try:
            outputs = json.loads(match.group(1))
        except Exception as exc:
            cabstat.declare_failure(StimelaCabOutputError(f"error parsing output dict from \"{self.name}'\"", exc))
        cabstat.declare_outputs(outputs)
        return output, None


# build dictionary of all available wrangler action classes

all_wranglers = {
    cls.specifier: cls for _, cls in vars().items()
        if isinstance(cls, type) and issubclass(cls, _BaseWrangler) and cls.specifier is not None
}






from typing import Dict, List, Optional, Any
from enum import Enum
from omegaconf import OmegaConf
from dataclasses import dataclass
import time
import yaml
import json
import secrets
import getpass
import logging
import time
import pwd
import grp
import os

import stimela
from scabha.basetypes import ( EmptyDictDefault, DictDefault, EmptyListDefault, 
                            ListDefault, EmptyClassDefault)
from stimela.exceptions import BackendError

session_id = secrets.token_hex(8)
session_user = getpass.getuser()

resource_labels = dict(stimela_user=session_user,
                    stimela_session_id=session_id)

try:
    import kubernetes
    from kubernetes.client import CustomObjectsApi
    from kubernetes.client.api import core_v1_api
    AVAILABLE = True
    STATUS = "ok"
except ImportError:
    AVAILABLE = False
    STATUS = "please reinstall with the optional kube dependency (stimela[kube])"

    def run(*args, **kw):
        raise RuntimeError(f"kubernetes backend {STATUS}")
    


# dict of methods for converting an object to text format
InjectedFileFormatters = dict(
    yaml = yaml.dump,
    json = json.dumps,
    txt = str
)

InjectedFileFormats = Enum("InjectedFileFormats", " ".join(InjectedFileFormatters.keys()), module=__name__)

@dataclass
class KubePodSpec(object):
    """Pod spec options. Used for job pods and for dask cluster pods."""
    @dataclass
    class PodLimits(object):
        """Pod limits and requuirements"""
        request: Optional[str] = None
        limit: Optional[str] = None

    # selects a specific pod type from a KubeBackendOptions.predefined_pod_specs
    type:           Optional[str] = None
    # memory limit/requirement
    memory:         Optional[PodLimits] = EmptyClassDefault(PodLimits)
    cpu:            Optional[PodLimits] = EmptyClassDefault(PodLimits)
    # arbitrary additional structure copied into the pod spec
    custom_pod_spec:  Dict[str, Any] = EmptyDictDefault()

@dataclass
class KubeBackendOptions(object):
    """
    Kube backend options class. Note that this can be defined globally in options, and 
    redefined/augmented at cab and step level.
    """

    @dataclass
    class Infrastructure(object):
        """Infrastructural options. These can only be defined globally -- ignored at cab/step level"""
        @dataclass
        class ExitOptions(object):
            cleanup_pods: bool = True                 # extra pod cleanup at exit
        @dataclass
        class StartupOptions(object):
            report_pods: bool = True                  # report any running pods for this user on startup
            cleanup_pods: bool = True                 # clean up any running pods on startup
            report_pvcs: bool = True                  # report any transient PVCs
            cleanup_pvcs: bool = True                 # cleanup any transient PVCs

        on_exit:    ExitOptions = EmptyClassDefault(ExitOptions)         # startup behaviour options
        on_startup: StartupOptions = EmptyClassDefault(StartupOptions)   # cleanup behaviour options

    @dataclass 
    class Volume(object):
        """Persistent volume claim config."""
        name: Optional[str] = None                                # populated with PVC name when allocated
        capacity: Optional[str] = None                            # capacity (e.g. 120Gi)
        storage_class_name: Optional[str] = None                  # k8s storage class
        access_modes: List[str] = ListDefault("ReadWriteOnce")    # ReadOnlyMany/ReadWriteMany for multi-attach
        provision_timeout: int = 1200                             # How long to wait for provisioning before timing out
        mount: Optional[str] = None                               # mount point

        from_snapshot: Optional[str] = None                       # create from snapshot

        # Status of PVC at start of sesssion or at start of step:
        # must_exist: reuse, error if it doesn't exist
        # allow_reuse: reuse if exists, else create
        # recreate: delete if exists and recreate
        # cant_exist: report an error if it exists, else create
        ExistPolicy = Enum("ExistPolicy", "must_exist allow_reuse recreate cant_exist", module=__name__)
        at_start: ExistPolicy = ExistPolicy.allow_reuse
        at_step: ExistPolicy = ExistPolicy.allow_reuse

        # commands issued in the volume at initialization. E.g. "chmod 777", "mkdir xxx", etc. These run as root.
        init_commands: List[str] = EmptyListDefault()
        # commands issued in the volume before each step initialization. E.g. "rm -fr *", "mkdir xxx", etc. 
        # These run as the user.
        step_init_commands: List[str] = EmptyListDefault()

        # lifecycle policy
        # persist: leave the PVC in place for future re-use
        # session: delete at end of stimela run
        # step: delete at end of step (only applies to per-step PVCs)
        Lifecycle = Enum("Lifecycle", "persist session step", module=__name__)
        lifecycle: Lifecycle = Lifecycle.session 

        append_id: bool = True    # for session- or step-lifecycle PVCs, append ID to name

        def __post_init__ (self):
            self.status = "Created"
            self.creation_time = time.time()
            self.metadata = None
            self.owner = None
            self.initialized = False

    # subclasses for options
    @dataclass
    class DaskCluster(object):
        enable: bool = False
        capture_logs: bool = True
        capture_logs_style: Optional[str] = "blue"
        name: Optional[str] = None
        num_workers: int = 1
        threads_per_worker: int = 1
        memory_limit: Optional[str] = None
        worker_pod: KubePodSpec = EmptyClassDefault(KubePodSpec)
        scheduler_pod: KubePodSpec = EmptyClassDefault(KubePodSpec)
        forward_dashboard_port: int = 8787          # set to non-0 to forward the http dashboard to this local port

    @dataclass
    class FileInjection(object):
        format: InjectedFileFormats = "txt"
        content: Any = ""

    @dataclass
    class LocalMount(object):
        path: str
        dest: str = ""              # destination path -- same as local if empty
        readonly: bool = False      # mount as readonly, but it doesn't work (yet?)
        mkdir: bool = False         # create dir, if it is missing


    enable: bool = True

    # infrastructure settings are global and can't be changed per cab or per step
    infrastructure: Infrastructure = EmptyClassDefault(Infrastructure)

    context:        Optional[str] = None   # k8s context -- use default if not given -- can't change
    namespace:      Optional[str] = None   # k8s namespace
    
    dask_cluster:   Optional[DaskCluster] = EmptyClassDefault(DaskCluster)  # if set, a DaskJob will be created
    service_account: str = "compute-runner"
    kubectl_path:   str = "kubectl"

    volumes:        Dict[str, Volume] = EmptyDictDefault()

    # inject_files:   Dict[str, FileInjection] = EmptyDictDefault()
    # pre_commands:   List[str] = EmptyListDefault()
    # local_mounts:   Dict[str, LocalMount] = EmptyDictDefault()
    env:            Dict[str, str] = EmptyDictDefault()
    dir:            Optional[str] = None                 # change to specific directory inside container

    provisioning_timeout: int = 600                      # timeout in seconds for pods/jobs to provision. 0 to disable

    connection_timeout: int = 60                         # connection timeout when talking to cluster, in seconds

    always_pull_images: bool = False                     # change to True to repull

    @dataclass
    class DebugOptions(object):
        verbose: int = 0                      # debug log level. Higher numbers mean more verbosity
        pause_on_start: bool = False        # pause instead of running payload
        pause_on_cleanup: bool = False      # pause before attempting cleanup

        save_spec: Optional[str] = None     # if set, pod/job specs will be saved as YaML to the named file. {}-substitutions apply to filename.

        # if >0, events will be collected and reported
        log_events:  bool = False
        # format string for reporting kubernetes events, this can include rich markup
        event_format:  str = "=NOSUBST('\\[k8s event type: {event.type}, reason: {event.reason}] {event.message}')"
        event_colors:  Dict[str, str] = DictDefault(
                                warning="blue", error="yellow", default="grey50")
    
    debug: DebugOptions = EmptyClassDefault(DebugOptions)


    job_pod: KubePodSpec = EmptyClassDefault(KubePodSpec)
    
    capture_logs_style: Optional[str] = "blue"

    @dataclass 
    class UserInfo(object):
        # user and group names and IDs -- if None, use local user
        name:           Optional[str] = None
        group:          Optional[str] = None
        uid:            Optional[int] = None
        gid:            Optional[int] = None
        gecos:          Optional[str] = None
        home:           Optional[str] = None     # home dir inside container, default is /home/{user}
        home_ramdisk:   bool = True              # home dir mounted as RAM disk, else local disk
        inject_nss:     bool = True              # inject user info for NSS_WRAPPER

    user: UserInfo = EmptyClassDefault(UserInfo)
    
    # user-defined set of pod types -- each is a pod spec structure keyed by pod_type
    predefined_pod_specs: Dict[str, Dict[str, Any]] = EmptyDictDefault()


KubeBackendSchema = OmegaConf.structured(KubeBackendOptions)
    
def is_available(opts: Optional[KubeBackendOptions]= None):
    return AVAILABLE

def get_status():
    return STATUS

def is_remote():
    return True

def init(backend: 'stimela.backend.StimelaBackendOptions', log: logging.Logger):
    from . import infrastructure
    global AVAILABLE, STATUS
    if not infrastructure.init(backend, log):
        AVAILABLE = False
        STATUS = "initialization error"

def close(backend: 'stimela.backend.StimelaBackendOptions', log: logging.Logger):
    from . import infrastructure
    if AVAILABLE:
        infrastructure.close(backend, log)

def cleanup(backend: 'stimela.backend.StimelaBackendOptions', log: logging.Logger):
    from . import infrastructure
    infrastructure.cleanup(backend, log)

def run(cab: 'stimela.kitchen.cab.Cab', params: Dict[str, Any], fqname: str,
        backend: 'stimela.backend.StimelaBackendOptions',
        log: logging.Logger, subst: Optional[Dict[str, Any]] = None,
        wrapper: Optional['stimela.backends.runner.BackendWrapper'] = None):
    from . import run_kube
    return run_kube.run(cab=cab, params=params, fqname=fqname, backend=backend, log=log, subst=subst)

_kube_config = _kube_context = _kube_namespace = None 

def get_kube_api(context: Optional[str]=None):
    global _kube_config, _kube_context, _kube_namespace

    if _kube_config is None:
        _kube_config = True
        kubernetes.config.load_kube_config(context=context)
        contexts, current_context = kubernetes.config.list_kube_config_contexts()
        if context is None:
            context = current_context['name']
        _kube_context = context
        for ctx in contexts:
            if ctx['name'] == context:
                _kube_namespace = ctx['context']['namespace']
                break
        else:
            _kube_namespace = "default"

    elif context is not None and context != _kube_context:
        raise BackendError(f"k8s context has changed (was {_kube_context}, now {context}), this is not permitted")

    return _kube_namespace, core_v1_api.CoreV1Api(), CustomObjectsApi()

def get_context_namespace():
    return _kube_context, _kube_namespace

_uid = os.getuid()
_gid = os.getgid()

session_user_info = KubeBackendOptions.UserInfo(
    name=session_user,
    group=grp.getgrgid(_gid).gr_name,
    uid=_uid,
    gid=_gid,
    home=f"/home/{session_user}",
    gecos=pwd.getpwuid(_uid).pw_gecos
)



import re
import logging
from typing import Optional, Any, Union, Dict, List
from dataclasses import dataclass

import stimela
from scabha.basetypes import EmptyListDefault
from scabha.exceptions import SubstitutionError
from stimela.kitchen.cab import Cab
from stimela.kitchen import wranglers
from scabha.substitutions import substitutions_from

from . import _CallableFlavour
from .python_flavours import format_dict_as_function_call


@dataclass
class CasaTaskFlavour(_CallableFlavour):
    kind: str = "casa-task"
    path: Optional[str] = None                       # path to CASA executable
    opts: Optional[List[str]] = EmptyListDefault()   # additional options 
    wrapper: Optional[str] = None                    # wrapper command (e.g. xvfb-run -a)

    def finalize(self, cab: Cab):
        super().finalize(cab)

        # catch CASA error messages, except the MeasTable::dUTC complaints which are all-pervasive
        rex = "(?P<content>(\tSEVERE\\s+(?!([a-z]+::)?MeasTable::dUTC)|ABORTING|\\*\\*\\* Error \\*\\*\\*)(.*))$"
        err_patt = re.compile(rex)
        cab._wranglers.append((
            err_patt, [
                wranglers.DeclareError(err_patt, "ERROR", message="CASA error: {content}" )
            ]
        ))

    def get_image_name(self, cab: Cab, backend: 'stimela.backend.StimelaBackendOptions'):
        from stimela import CONFIG
        from stimela.backends import resolve_image_name
        return resolve_image_name(backend, cab.image or CONFIG.images['default-casa'])

    def get_arguments(self, cab: Cab, params: Dict[str, Any], subst: Dict[str, Any], 
                            virtual_env: Optional[str]=None, 
                            check_executable: bool = True,
                            log: Optional[logging.Logger] = None):

        with substitutions_from(subst, raise_errors=True) as context:
            try:
                command = context.evaluate(cab.command, location=["command"])
            except Exception as exc:
                raise SubstitutionError(f"error substituting casa task '{cab.command}'", exc)
            casa_config = stimela.CONFIG.opts.runtime.get('casa', {})
            casa = self.path or casa_config.get('path', 'casa')
            try:
                casa = context.evaluate(casa, location=["path"])
            except Exception as exc:
                raise SubstitutionError(f"error substituting casa path '{casa}'", exc)
            casa_opts = self.opts or casa_config.get('opts', ["--log2term", "--nologger", "--nologfile"]) or []
            if casa_opts:
                try:
                    casa_opts = [context.evaluate(opt, location=["opts"]) for opt in casa_opts]
                except Exception as exc:
                    raise SubstitutionError(f"error substituting casa options '{casa_opts}'", exc)
            wrapper = self.wrapper if self.wrapper is not None else casa_config.get('wrapper', 'xvfb-run -a')
            if wrapper:
                try:
                    wrapper = context.evaluate(wrapper, location=["wrapper"]).split()
                except Exception as exc:
                    raise SubstitutionError(f"error substituting wrapper '{wrapper}'", exc)
            else:
                wrapper = []

        self.command_name = command
        pass_params = dict(cab.filter_input_params(params))

        # log invocation
        if log:
            log.info(f"preparing CASA task call:", extra=dict(prefix="###", style="dim"))
            for line in format_dict_as_function_call(command, pass_params, indent=4):
                log.info(f"    {line}", extra=dict(prefix="###", style="dim"))

        # parse the params direcly as python dictionary
        # no need to string conversion between strings/bytes/unicode
        # this works for both python 2.7 and 3.x
        code = f"{command}(**{pass_params})"

        args =  wrapper + casa.strip().split() + list(casa_opts)
        
        return args + ["-c", code], args + ["-c", "..."]
    





from .test_recipe import change_test_dir, run, verify_output


def test_param_file_input():
    print("===== expecting no errors =====")
    retcode, output = run("stimela -b native run test_param_file.yml test-param-file -pf param_file.yml")
    assert retcode == 0
    print(output)
    assert verify_output(output, "A string!", "['A', 'list', '!']")
    assert not verify_output(output, "gremlin")
    assert verify_output(output, "Found a unicorn!")



import logging
import atexit
import json
import time
from typing import Optional, Dict, List

from stimela.backends import StimelaBackendOptions
from stimela.stimelogging import log_exception
from stimela.task_stats import update_process_status

from stimela.exceptions import BackendError
from . import session_id, session_user, resource_labels, run_kube, KubeBackendOptions, get_kube_api, get_context_namespace

Lifecycle = KubeBackendOptions.Volume.Lifecycle

from kubernetes import client
from kubernetes.client.rest import ApiException
from kubernetes.config import ConfigException

from .kube_utils import resolve_unit

# Note that transient PVCs have a stimela claim name (without ID suffix) and an actual name (with ID suffix)
# Non-transient PVCs have the same name (no suffix in each case)
# This maps stimela claim names to Volume objects
active_pvcs: Dict[str, KubeBackendOptions.Volume] = {}

# This is a dict of terminating PVS, mapping actual names to stimela claim names
terminating_pvcs: Dict[str, str]

# This is a dict of PVCs requiring initialization
session_init_commands: Dict[str, List[str]] = {}

# logger used for global kube messages
klog: Optional[logging.Logger] = None

def _delete_pod(kube_api, podname, namespace, log, warn_not_found=True):
    log.info(f"deleting pod {podname}")
    try:
        resp = kube_api.delete_namespaced_pod(name=podname, namespace=namespace)
        log.debug(f"delete_namespaced_pod({podname}): {resp}")
    except ApiException as exc:
        body = json.loads(exc.body)
        if "reason" in body and body["reason"] == "NotFound" and warn_not_found:
            log.warning(f"pod {podname} not found, this is probably OK, perhaps it just died on its own")
        else:
            log_exception(BackendError(f"k8s API error while deleting pod {podname}", (exc, body)), 
                        severity="error", log=log)

def cleanup(backend: StimelaBackendOptions, log: logging.Logger):
    return init(backend, log, cleanup=True)

def init(backend: StimelaBackendOptions, log: logging.Logger, cleanup: bool = False):
    global klog
    klog = log.getChild("kube")
    kube = backend.kube

    if cleanup:
        klog.info("cleaning up backend")
    else:
        atexit.register(close, backend, klog)
        klog.info("initializing kube backend")

    try:
        namespace, kube_api, _ = get_kube_api(kube.context)
    except ConfigException as exc:
        log_exception(exc, log=klog)
        log_exception(BackendError("error initializing kube backend", exc), log=klog)
        return False
    
    context, namespace = get_context_namespace()
    klog.info(f"k8s context is {context}, namespace is {namespace}")

    if cleanup or kube.infrastructure.on_startup.report_pods or kube.infrastructure.on_startup.cleanup_pods:
        klog.info("checking for k8s pods from other sessions")

        try:
            pods = kube_api.list_namespaced_pod(namespace=namespace, 
                                                label_selector=f"stimela_user={session_user}")
        except ApiException as exc:
            raise BackendError(f"k8s API error while listing pods", json.loads(exc.body))
        
        running_pods = []
        for pod in pods.items:
            if pod.status.phase in ("Running", "Pending") and not pod.metadata.deletion_timestamp:
                running_pods.append(pod.metadata.name)

        if running_pods:
            if cleanup or kube.infrastructure.on_startup.cleanup_pods:
                klog.warning(f"you have {len(running_pods)} pod(s) running from another stimela session")
                if not cleanup:
                    klog.warning(f"since kube.infrastructure.on_starup.cleanup_pods is set, these will be terminated")
                for podname in running_pods:
                    _delete_pod(kube_api, podname, namespace, klog)

            elif kube.infrastructure.on_startup.report_pods:
                klog.warning(f"you have {len(running_pods)} pod(s) running from another stimela session")
                klog.warning(f"set kube.infrastructure.on_startup.report_pods to false to disable this warning")
        else:
            klog.info("no pods running")

    refresh_pvc_list(kube)

    # cleanup transient PVCs
    transient_pvcs = {name: pvc for name, pvc in active_pvcs.items() 
                        if pvc.metadata.labels and 'stimela_transient_pvc' in pvc.metadata.labels}
    
    if transient_pvcs:
        if cleanup or kube.infrastructure.on_startup.cleanup_pvcs:
            klog.warning(f"you have {len(transient_pvcs)} transient PVC(s) from another stimela session")
            if not cleanup:
                klog.warning(f"since kube.infrastructure.on_starup.cleanup_pvcs is set, these will be deleted")
            delete_pvcs(kube, transient_pvcs.keys(), log=log, refresh=False, force=True)
            refresh_pvc_list(kube)

        elif kube.infrastructure.on_startup.report_pvcs:
            klog.warning(f"you have {len(transient_pvcs)} transient PVC(s) from another stimela session")
            klog.warning(f"set kube.infrastructure.on_startup.report_pvcs to false to disable this warning")
    elif cleanup:
        klog.info("checking for transient PVCs: none found")

    # resolve global-level volumes
    if not cleanup and kube.volumes:
        resolve_volumes(kube, log=klog, refresh=False) # no refresh needed

    return True


def refresh_pvc_list(kube: KubeBackendOptions):
    namespace, kube_api, _ = get_kube_api()
    global active_pvcs, terminating_pvcs
    # get existing pvcs
    try:
        list_pvcs = kube_api.list_namespaced_persistent_volume_claim(namespace)
    except ApiException as exc:
        raise BackendError(f"k8s API error while listing PVCs", json.loads(exc.body)) from None
    pvc_names = []
    terminating_pvcs = {}
    # convert to PVC entry 
    for pvc in list_pvcs.items:
        if pvc.metadata.labels and 'stimela_pvc_name' in pvc.metadata.labels:
            name = pvc.metadata.labels['stimela_pvc_name']
        else:
            name = pvc.metadata.name
        # add to terminating list, if marked for deletion
        if pvc.metadata.deletion_timestamp:
            terminating_pvcs[pvc.metadata.name] = name
            continue
        pvc_names.append(name)
        # insert new entry if it doesn't exist
        if name not in active_pvcs:
            active_pvcs[name] = KubeBackendOptions.Volume(name=pvc.metadata.name, 
                                    capacity=pvc.spec.resources.requests['storage'], 
                                    lifecycle=Lifecycle.persist)
        active_pvcs[name].status = pvc.status.phase
        active_pvcs[name].metadata = pvc.metadata
        active_pvcs[name].owner = pvc.metadata.labels and pvc.metadata.labels.get("stimela_user")
    # delete stale entries
    active_pvcs = {name: active_pvcs[name] for name in pvc_names} 


def resolve_volumes(kube: KubeBackendOptions, log: logging.Logger, step_token=None, refresh=True):
    namespace, kube_api, _ = get_kube_api()
    ExistsPolicy = KubeBackendOptions.Volume.ExistPolicy
    global terminating_pvcs

    if refresh:
        refresh_pvc_list(kube)

    # look for required PVCs
    for name, pvc in kube.volumes.items():
        exist_policy = pvc.at_step if step_token else pvc.at_start
        exist_policy_desc = 'at step' if step_token else 'at start'
        # Exists? Check that size is enough
        pvc0 = active_pvcs.get(name)
        # check for existing PVCs
        if pvc0 is not None:
            if exist_policy == ExistsPolicy.cant_exist:
                raise BackendError(f"PVC '{name}' already exists: according to its '{exist_policy_desc}' policy, this is an error")
            # check if we need to re-create
            if exist_policy == ExistsPolicy.recreate:
                log.info(f"PVC '{name}' already exists: re-creating according to its '{exist_policy_desc}' policy")
                delete_pvcs(kube, pvc_names=[name], log=log, force=True, refresh=False)
                pvc0 = None
            # else reusing -- check capacity 
            elif pvc.capacity is None or resolve_unit(pvc.capacity) <= resolve_unit(pvc0.capacity):
                log.info(f"found existing PVC '{name}' of size {pvc0.capacity}, status is {pvc0.status}")
                # copy name -- pre-existsing PVC may have been auto-named
                pvc.name = pvc0.name
            else:
                raise BackendError(f"Existing PVC '{name}' of size {pvc0.capacity} is smaller than the requested {pvc.capacity}")
            
        # Doesn't exist? Create
        if pvc0 is None:
            if exist_policy == ExistsPolicy.must_exist:
                raise BackendError(f"PVC '{name}' doesn't exist: according to its '{exist_policy_desc}' policy, this is an error")
            if pvc.storage_class_name is None or pvc.capacity is None:
                raise BackendError(f"Can't create PVC '{name}': storage class name or capacity not specified")
            # create new one
            pvc.name = name
            pvc.status = 'Creating'
            pvc.creation_time = time.time()
            labels = resource_labels.copy()
            labels['stimela_pvc_name'] = name
            labels['stimela_pvc_initialized'] = ''
            # append token for limited-lifecycle PVCs
            if pvc.append_id:
                if pvc.lifecycle == Lifecycle.session:
                    pvc.name = f"{name}-{session_id}"
                    labels['stimela_transient_pvc'] = 'session'
                elif pvc.lifecycle == Lifecycle.step:
                    if step_token is None:
                        raise BackendError(f"PVC '{name}' with lifecycle=step not allowed at infrastructure level")
                    pvc.name = f"{name}-{step_token}"
                    labels['stimela_transient_pvc'] = 'step'
            # if existing PVC with that is still terminating, wait
            if pvc.name in terminating_pvcs:
                log.info(f"waiting for existing PVC '{pvc.name}' to terminate before re-creating")
                _await_pvc_termination(namespace, pvc, log=log)
            # create
            newpvc = client.V1PersistentVolumeClaim()
            newpvc.metadata = client.V1ObjectMeta(name=pvc.name, labels=labels)
            if pvc.from_snapshot:
                data_source = dict(name=pvc.from_snapshot, 
                                   kind='VolumeSnapshot',
                                   apiGroup='snapshot.storage.k8s.io')
                log.info(f"creating new PVC '{pvc.name}' of size {pvc.capacity} from snapshot '{pvc.from_snapshot}'")
            else:
                log.info(f"creating new PVC '{pvc.name}' of size {pvc.capacity}")
                data_source = None
            newpvc.spec = client.V1PersistentVolumeClaimSpec(
                access_modes=list(pvc.access_modes),
                storage_class_name=pvc.storage_class_name,
                data_source=data_source,
                resources=client.V1ResourceRequirements(requests={"storage": pvc.capacity}))
            try:
                resp = kube_api.create_namespaced_persistent_volume_claim(namespace, newpvc)
            except ApiException as exc:
                raise BackendError(f"k8s API error while creating PVC '{pvc.name}'", json.loads(exc.body)) from None
            pvc.owner = session_user
            active_pvcs[name] = pvc
        # reusing volume -- but check if it is initialized
        else:
            pvc0.initialized = pvc0.metadata.labels and pvc0.metadata.labels.get("stimela_pvc_initialized")

    return list(kube.volumes.keys())


def await_pvcs(namespace, pvc_names, log: logging.Logger):
    namespace, kube_api, _ = get_kube_api()

    waiting_pvcs = set(pvc_names)
    waiting_reported = set()

    while waiting_pvcs:
        update_process_status()
        # recheck state of all PVCs we're waiting on
        for name in list(waiting_pvcs):
            pvc = active_pvcs.get(name)
            if pvc is None:
                raise BackendError(f"'{name}' does not refer to a previously defined PVC")
            if pvc.status == "Bound":
                waiting_pvcs.remove(name)
                continue
            # get updated status
            try:
                pvc_entry = kube_api.read_namespaced_persistent_volume_claim(name=pvc.name, namespace=namespace)
            except ApiException as exc:
                raise BackendError(f"k8s API error while reading PVC '{pvc.name}'", json.loads(exc.body))
            pvc.status = pvc_entry.status.phase
            if pvc.status == 'Bound':
                log.info(f"PVC '{pvc.name}' is now bound")
                waiting_pvcs.remove(name)
                continue
            # check for timeout
            if pvc.provision_timeout and time.time() > pvc.creation_time + pvc.provision_timeout:
                raise BackendError(f"timed out waiting for PVC '{pvc.name}' to provision")
            # report that we're waiting
            if name not in waiting_reported:
                log.info(f"waiting for PVC '{pvc.name}' to provision")
                waiting_reported.add(name)
        # wait some more
        time.sleep(1)

def _await_pvc_termination(namespace, pvc: KubeBackendOptions.Volume, log: logging.Logger):
    global terminating_pvcs
    namespace, kube_api, _ = get_kube_api()
    time0 = time.time()
    while True:
        update_process_status()
        # recheck state of all PVCs we're waiting on
        try:
            kube_api.read_namespaced_persistent_volume_claim(name=pvc.name, namespace=namespace)
        except ApiException as exc:
            body = json.loads(exc.body)
            if "reason" in body and body["reason"] == "NotFound":
                log.info(f"PVC '{pvc.name}' has terminated")
                del terminating_pvcs[pvc.name]
                return
            else:
                raise BackendError(f"k8s API error while reading PVC '{pvc.name}'", json.loads(exc.body))
        if pvc.provision_timeout and time.time() > time0 + pvc.provision_timeout:
            raise BackendError(f"timed out waiting for PVC '{pvc.name}' to terminate")
        # wait some more
        time.sleep(1)



def delete_pvcs(kube: KubeBackendOptions, pvc_names, log: logging.Logger, force=False, step=True, session=False, refresh=True):
    namespace, kube_api, _ = get_kube_api()
    global terminating_pvcs

    if refresh:
        refresh_pvc_list(kube)
    
    for name in pvc_names:
        if name in terminating_pvcs:
            continue

        pvc = active_pvcs.get(name)
        if pvc is None:
            raise BackendError(f"'{name}' does not refer to a previously defined PVC")
        
        if pvc.status != 'Terminating' and \
            force or \
            (step and pvc.lifecycle == Lifecycle.step) or \
            (session and pvc.lifecycle == Lifecycle.session):
            log.info(f"deleting PVC '{pvc.name}'")
            try:
                resp = kube_api.delete_namespaced_persistent_volume_claim(name=pvc.name, namespace=namespace)
            except ApiException as exc:
                body = json.loads(exc.body)
                log_exception(BackendError(f"k8s API error while deleting PVC '{pvc.name}'", (exc, body)), 
                                severity="error", log=log)
                continue
            log.debug(f"delete_namespaced_persistent_volume_claim({pvc.name}): {resp}")
            pvc.status = 'Terminating'
            terminating_pvcs[pvc.name] = name


def close(backend: StimelaBackendOptions, log: logging.Logger):
    kube = backend.kube
    context, namespace = get_context_namespace()
    if context is None:
        return 
    
    klog.info("closing kube backend")

    # release PVCs
    delete_pvcs(kube, list(active_pvcs.keys()), log=klog, session=True, step=True, refresh=False)

    # cleanup pods, if any
    if kube.infrastructure.on_exit.cleanup_pods:
        namespace, kube_api, _ = run_kube.get_kube_api() 

        try:
            pods = kube_api.list_namespaced_pod(namespace=namespace, 
                                                label_selector=f"stimela_session_id={session_id}")
        except ApiException as exc:
            body = json.loads(exc.body)
            log_exception(BackendError(f"k8s API error while listing pods", (exc, body)), severity="error", log=klog)
            return
        
        running_pods = []
        for pod in pods.items:
            if pod.status.phase in ("Running", "Pending") and not pod.metadata.deletion_timestamp:
                running_pods.append(pod.metadata.name)

        if running_pods:
            klog.warning(f"you have {len(running_pods)} pod(s) still pending or running from this session")
            klog.warning(f"since kube.infrastructure.on_exit.cleanup_pods is set, these will be terminated")
            for podname in running_pods:
                _delete_pod(kube_api, podname, namespace, klog)

    atexit.unregister(close)




import os.path
import sys
import glob
import hashlib
import pathlib
import dill as pickle

import uuid
from dataclasses import make_dataclass

from omegaconf.omegaconf import OmegaConf, DictConfig, ListConfig
from omegaconf.errors import OmegaConfBaseException
from typing import Any, List, Dict, Optional, OrderedDict, Union, Callable

from scabha.exceptions import ScabhaBaseException

from yaml.error import YAMLError
from .deps import ConfigDependencies
from .resolvers import resolve_config_refs
from .common import *

# path for cache
CACHEDIR = os.environ.get("CONFIGURATT_CACHE_DIR") or os.path.expanduser(
    "~/.cache/configuratt")


def _compute_hash(filelist, extra_keys):
    filelist = list(filelist) + list(extra_keys)
    return hashlib.md5(" ".join(filelist).encode()).hexdigest()


def set_cache_dir(cachedir: str):
    global CACHEDIR
    CACHEDIR = cachedir


def clear_cache(log=None):
    if os.path.isdir(CACHEDIR):
        files = glob.glob(f"{CACHEDIR}/*")
        log and log.info(f"clearing {len(files)} cached config(s) from cache")
    else:
        files = []
        log and log.info(f"no configs in cache")
    for filename in files:
        try:
            os.unlink(filename)
        except Exception as exc:
            log and log.error(f"failed to remove cached config {filename}: {exc}")
            sys.exit(1)

def load_cache(filelist: List[str], extra_keys=[], verbose=None):
    filehash = _compute_hash(filelist, extra_keys)
    if not os.path.isdir(CACHEDIR):
        if verbose:
            print(f"{CACHEDIR} does not exist")
        return None, None
    filename = os.path.join(CACHEDIR, filehash)
    if not os.path.exists(filename):
        if verbose:
            print(f"hash file {filename} does not exist")
        return None, None
    # check that all configs are older than the cache
    cache_mtime = os.path.getmtime(filename)
    for f in filelist:
        if os.path.getmtime(f) > cache_mtime:
            if verbose:
                print(f"Config {f} is newer than the cache, forcing reload")
            return None, None
    # load cache
    try:
        conf, deps = pickle.load(open(filename, 'rb'))
        if not isinstance(deps, ConfigDependencies):
            raise TypeError(f"cached deps object is of type {type(deps)}, expecting ConfigDependencies")
    except Exception as exc:
        print(f"Error loading cached config from {filename}: {exc}. Removing the cache.")
        os.unlink(filename)
        return None, None
    # check that all dependencies are older than the cache
    if deps.have_deps_changed(cache_mtime, verbose=verbose):
        return None, None
    if verbose:
        print(f"Loaded cached config for {' '.join(filelist)} from {filename}")
    return conf, deps


def save_cache(filelist: List[str], conf, deps: ConfigDependencies, extra_keys=[], verbose=False):
    pathlib.Path(CACHEDIR).mkdir(parents=True, exist_ok=True)
    filelist = list(filelist)   # add self to dependencies
    filehash = _compute_hash(filelist, extra_keys)
    filename = os.path.join(CACHEDIR, filehash)
    # add ourselves to dependencies, so that cache is cleared if implementation changes
    deps.add(__file__, version=PACKAGE_VERSION)
    pickle.dump((conf, deps), open(filename, "wb"), 2)
    if verbose:
        print(f"Caching config for {' '.join(filelist)} as {filename}")




